"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to deleting Revit views. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Revit.Common import delete as rDel, parameter_get_utils as rParaGet
from duHast.Utilities.Objects import result as res
from duHast.Utilities import utility as util
from duHast.Revit.Views.views import get_views_not_on_sheet


def delete_views(doc, view_rules, collector_views):
    """
    Deletes views based on view rules supplied.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view_rules: Rules used to check whether a view should be deleted. Rules are based on parameters attached to views and their values.
    :type view_rules: list of lists, where each inner list has the format [parameter name, condition test method, value to test against]
    :param collector_views: A filtered element collector containing views.
    :type collector_views: Autodesk.Revit.DB.FilteredElementCollector

    :return: Result class instance.
        - `result` (bool): True if all views were deleted, otherwise False.
        - `message` (str): Deletion status.
    :rtype: :class:`.Result`
    """

    ids = []
    view_counter = 0
    for v in collector_views:
        # filter out revision schedules '<', sheets and other view types which can not be deleted
        if (
            util.encode_ascii(v.Name)[0] != "<"
            and v.ViewType != rdb.ViewType.Internal
            and v.ViewType != rdb.ViewType.Undefined
            and v.ViewType != rdb.ViewType.ProjectBrowser
            and v.ViewType != rdb.ViewType.DrawingSheet
            and v.ViewType != rdb.ViewType.SystemBrowser
        ):
            view_counter = +1
            paras = v.GetOrderedParameters()
            rule_match = True
            for para_name, paraCondition, conditionValue in view_rules:
                for p in paras:
                    if p.Definition.Name == para_name:
                        rule_match = rule_match and rParaGet.check_parameter_value(
                            p, paraCondition, conditionValue
                        )
            if rule_match == True:
                # delete view
                ids.append(v.Id)
    # make sure we are not trying to delete all views (this allowed when a model is opened into memory only, but that model will crash when trying to open into UI)
    if len(ids) == view_counter and len(ids) > 0:
        ids.pop()
    # delete all views at once
    result = rDel.delete_by_element_ids(
        doc, ids, "deleting views not matching filters", "views"
    )
    return result


def delete_views_not_on_sheets(doc, filter):
    """
    Deletes all views not placed on sheets, including schedules and legends matching the given filter.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param filter: A function that checks whether a view should be deleted.
    :type filter: function(view) -> bool

    :return: Result class instance.
        - `result` (bool): True if all views were deleted, otherwise False.
        - `message` (str): Deletion status.
    :rtype: :class:`.Result`
    """

    ids = []
    return_value = res.Result()
    views_not_on_sheets = get_views_not_on_sheet(doc)
    for view_not_on_sheet in views_not_on_sheets:
        if filter(view_not_on_sheet):
            ids.append(view_not_on_sheet.Id)
    # check we are not trying to delete all views
    if len(views_not_on_sheets) == len(ids) and len(views_not_on_sheets) > 0:
        # remove a random view from this list
        ids.pop(0)
    if len(ids) > 0:
        return_value = rDel.delete_by_element_ids(
            doc,
            ids,
            "deleting " + str(len(views_not_on_sheets)) + " views not on sheets",
            "views",
        )
    else:
        return_value.update_sep(True, "No views not placed on sheets found.")
    return return_value


def delete_unused_elevation_view_markers(doc):
    """
    Deletes all unused elevation markers (no elevation is created by the marker).

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: Result class instance.
        - `result` (bool): True if all unused elevation markers were deleted, otherwise False.
        - `message` (str): Deletion status.
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    ele = rdb.FilteredElementCollector(doc).OfClass(rdb.ElevationMarker)
    # items to be deleted
    ids = []
    # set up view counter (how many views will be deleted)
    counter = 0
    # loop over markers
    for e in ele:
        # check if view count is 0 (unused marker)
        if e.CurrentViewCount == 0:
            # add to list of views to be deleted
            ids.append(e.Id)
            counter += 1
    if len(ids) > 0:
        return_value = rDel.delete_by_element_ids(
            doc, ids, "deleting unused view markers: " + str(counter), "view marker"
        )
    else:
        return_value.update_sep(True, "No unused elevation markers in model")
    return return_value


def delete_sheets(doc, view_rules, collector_views):
    """
    Deletes sheets based on rules.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view_rules: A set of rules. If view matches rule it will be deleted.
    :type view_rules: array in format [parameter name, condition test method, value to test against
    :param collector_views: A filtered element collector containing view instances.
    :type collector_views: Autodesk.Revit.DB.FilteredElementCollector
    :return:
        Result class instance.
        - .status True if all sheets matching filter where deleted. Otherwise False.
        - .message will contain deletion status.
    :rtype: :class:`.Result`
    """

    ids = []
    for v in collector_views:
        if v.ViewType == rdb.ViewType.DrawingSheet:
            paras = v.GetOrderedParameters()
            rule_match = True
            for para_name, paraCondition, conditionValue in view_rules:
                for p in paras:
                    if p.Definition.Name == para_name:
                        rule_match = rule_match and rParaGet.check_parameter_value(
                            p, paraCondition, conditionValue
                        )
            if rule_match == True:
                # delete view
                ids.append(v.Id)
    result = rDel.delete_by_element_ids(doc, ids, "deleting sheets", "sheets")
    return result


def delete_all_sheets(doc):
    """
    Deletes all sheets in a model

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return:
        Result class instance.
        - .status True if all sheets where deleted. Otherwise False.
        - .message will contain deletion status.
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    ids = []
    collector_sheets = rdb.FilteredElementCollector(doc).OfClass(rdb.View)
    for v in collector_sheets:
        if v.ViewType == rdb.ViewType.DrawingSheet:
            ids.append(v.Id)
    if len(ids) > 0:
        return_value = rDel.delete_by_element_ids(
            doc, ids, "deleting all sheets", "sheets"
        )
    else:
        return_value.update_sep(True, "No sheets in the model")
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to purging unused image links.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Common import common as com
from duHast.Revit.Links.image_links import (
    get_image_instances_in_model,
    get_image_types_in_model,
)


def get_all_unused_image_type_ids_in_model(doc):
    """
    Gets all image types with no instances placed in a model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of ids representing image types.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    unused_images = com.get_not_placed_types(
        doc, get_image_types_in_model, get_image_instances_in_model
    )
    unused_type_ids = []
    for i in unused_images:
        unused_type_ids.append(i.Id)
    return unused_type_ids


def get_all_unused_image_type_ids_in_model_with_group_check(doc):
    """
    Gets all image types with no instance placed in a model but includes group definition check.
    This only returns valid data if at least one instance of the group is placed in the model!!
    Otherwise images in groups which are not placed will not be flagged by this filter!
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of ids representing image types.
    :rtype: list  Autodesk.Revit.DB.ElementId
    """

    unused_type_ids = get_all_unused_image_type_ids_in_model(doc)
    # and filter by any type id's in groups which may not be placed and hence no instance present in the model
    unused_type_ids = com.get_unused_type_ids_from_detail_groups(doc, unused_type_ids)
    return unused_type_ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit file operations utility functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from Autodesk.Revit.DB import (
    RelinquishOptions,
    TransactWithCentralOptions,
    SaveAsOptions,
    SimpleWorksetConfiguration,
    SynchronizeWithCentralOptions,
    WorksharingUtils,
    WorksharingSaveAsOptions,
)

from duHast.Utilities import files_io as fileIO
from duHast.Utilities.Objects import result as res


def sync_file(
    doc, compact_central_file=False  # type: bool
):
    # type: (...) -> res.Result
    """
    Synchronizes a Revit central file.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param compact_central_file: option to compact the central file, defaults to False
    :type compact_central_file: bool, optional
    :return:
        Result class instance.
        - .status True if successfully synced file. Otherwise False.
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # set up sync settings
    ro = RelinquishOptions(True)
    transaction_options = TransactWithCentralOptions()
    sync = SynchronizeWithCentralOptions()
    sync.Comment = "Synchronized by Revit Batch Processor"
    sync.Compact = compact_central_file
    sync.SetRelinquishOptions(ro)
    # Synch it
    try:
        # save local first ( this seems to prevent intermittent crash on sync(?))
        doc.Save()
        doc.SynchronizeWithCentral(transaction_options, sync)
        # relinquish all
        WorksharingUtils.RelinquishOwnership(doc, ro, transaction_options)
        return_value.message = "Successfully synched file."
    except Exception as e:
        return_value.update_sep(False, "Failed with exception: {}".format(e))
    return return_value


def saves_as_workshared_file(
    doc, full_file_name  # type: str
):
    # type: (...) -> res.Result
    """
    Saves a Revit project file as a workshared file.
    Save as options are:
    Workset configuration is : Ask users on open to specify.
    Any existing file will be overwritten.
    Number of backups is 5
    File will bew compacted on save.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param full_file_name: The fully qualified file path of where to save the file.
    :type full_file_name: string
    :return:
        Result class instance.
        - .status True if successfully saved file, otherwise False.
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    try:
        worksharing_save_as_option = WorksharingSaveAsOptions()
        worksharing_save_as_option.OpenWorksetsDefault = (
            SimpleWorksetConfiguration.AskUserToSpecify
        )
        worksharing_save_as_option.SaveAsCentral = True
        save_option = SaveAsOptions()
        save_option.OverwriteExistingFile = True
        save_option.SetWorksharingOptions(worksharing_save_as_option)
        save_option.MaximumBackups = 5
        save_option.Compact = True
        doc.SaveAs(full_file_name, save_option)
        return_value.message = "Successfully saved file: " + str(full_file_name)
    except Exception as e:
        return_value.update_sep(False, "Failed with exception: {}".format(e))
    return return_value


def save_as_family(
    doc,
    target_directory_path,
    current_full_file_name,
    name_data,
    file_extension=".rfa",
    compact_file=False,
):
    """
    Saves a family file under new name in given location.

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param target_directory_path: The directory path of where to save the file.
    :type target_directory_path: str
    :param current_full_file_name: The current (old) name of the file.
    :type current_full_file_name: str
    :param name_data:  Old name and new name are Revit file names without file extension. Used to rename the family on save from old name to new name.
    :type name_data: List of string arrays in format[[oldname, newName]]
    :param file_extension: The file extension used for the new file, defaults to '.rfa'
    :type file_extension: str, optional
    :param compact_file: Flag whether family is to be compacted on save, defaults to False
    :type compact_file: bool, optional
    :return:
        Result class instance.
            - .status True if successfully saved file, otherwise False.
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    revit_file_name = fileIO.get_file_name_without_ext(current_full_file_name)
    new_file_name = ""
    match = False
    # find new file name in list past in
    for old_name, new_name in name_data:
        if revit_file_name.startswith(old_name):
            match = True
            return_value.message = "Found file name match for: {} new name: {}".format(
                revit_file_name, new_name
            )
            # save file under new name
            new_file_name = target_directory_path + "\\" + new_name + file_extension
            break
    if match == False:
        # save under same file name
        new_file_name = target_directory_path + "\\" + revit_file_name + file_extension
        return_value.message = "Found no file name match for: {}".format(
            current_full_file_name
        )
    try:
        # setup save as option
        so = SaveAsOptions()
        so.OverwriteExistingFile = True
        so.MaximumBackups = 5
        so.SetWorksharingOptions(None)
        so.Compact = compact_file
        doc.SaveAs(new_file_name, so)
        return_value.update_sep(True, "Saved file: {}".format(new_file_name))
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to save revit file to new location with exception: {}".format(e),
        )
    return return_value


def save_as(
    doc, target_directory_path, current_full_file_name, name_data, file_extension=".rvt"
):
    """
    Saves a project file under new name in given location.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param target_directory_path: The directory path of where to save the file.
    :type target_directory_path: str
    :param current_full_file_name: The current (old) name of the file.
    :type current_full_file_name: str
    :param name_data: Old name and new name are revit file names without file extension. Used to rename the model on save from old name to new name.
    :type name_data: List of string arrays in format[[oldname, newName]]
    :param file_extension: The file extension used for the new file, defaults to '.rvt'
    :type file_extension: str, optional
    :return:
        Result class instance.
        - .status True if successfully saved file, otherwise False.
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # added str() around this expression to satisfy sphinx auto code documentation
    # it will throw an exception when concatenating the string in the return statement
    revit_file_name = str(fileIO.get_file_name_without_ext(current_full_file_name))
    new_file_name = ""
    match = False
    for old_name, new_name in name_data:
        if revit_file_name.startswith(old_name):
            match = True
            return_value.message = "Found file name match for: {} new name: {}".format(
                revit_file_name, new_name
            )
            # save file under new name
            new_file_name = target_directory_path + "\\" + new_name + file_extension
            break
    if match == False:
        # save under same file name
        new_file_name = target_directory_path + "\\" + revit_file_name + file_extension
        # added str.format around this expression to satisfy sphinx auto code documentation
        return_value.message = "Found no file name match for: {}".format(
            current_full_file_name
        )
    try:
        return_value.status = saves_as_workshared_file(doc, new_file_name).status
        return_value.append_message("Saved file: {}".format(new_file_name))
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to save revit file to new location; {} with exception: ".format(
                new_file_name, e
            ),
        )
    return return_value


def save_file(
    doc, compact_file=False  # type: bool
):
    # type: (...) -> res.Result
    """
    Saves a non workshared Revit file. To be used for families and non workshared revit files only.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param compact_file: True file will be compacted on save, defaults to False
    :type compact_file: bool, optional
    :return:
            Result class instance.
            - .status True if file was saved successfully. Otherwise False.
            - .message = 'Saved revit file!'
            On exception:
            - result.status (bool) will be False.
            - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    try:
        save_options = SaveOptions()
        save_options.Compact = compact_file
        doc.Save(save_options)
        return_value.update_sep(True, "Saved revit file!")
    except Exception as e:
        return_value.update_sep(
            False, "Failed to save revit file with exception: {}".format(e)
        )
    return return_value


def enable_worksharing(
    doc,  #
    workset_name_grid_level="Shared Levels and Grids",  # type: str
    workset_name="Workset1",  # type: str
):
    # type: (...) -> res.Result
    """
    Enables worksharing in a non workshared revit project file.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param workset_name_grid_level: _description_, defaults to 'Shared Levels and Grids'
    :type workset_name_grid_level: str, optional
    :param workset_name: _description_, defaults to 'Workset1'
    :type workset_name: str, optional

    :return:
            Result class instance.
            - .status True if worksharing was enabled successfully. Otherwise False.
            - .message = 'Successfully enabled worksharing.'
            On exception:
            - result.status (bool) will be False.
            - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    try:
        doc.EnableWorksharing(workset_name_grid_level, workset_name)
        return_value.message = "Successfully enabled worksharing."
    except Exception as e:
        return_value.update_sep(False, "Failed with exception: {}".format(e))
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family shared parameters purge unused utilities.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This will delete all shared parameter definitions which are not used by any family parameter.

- requires a revit shared parameter processor object

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# class used for stats reporting
from duHast.Utilities.Objects import result as res
from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData
from duHast.Revit.SharedParameters import shared_parameter_data as rSharedParaData
from duHast.Revit.Categories import categories as rCat
from duHast.Revit.Common import delete as rDel

import Autodesk.Revit.DB as rdb


def purge_unused(doc, processor):
    """
    This will delete all shared parameter definitions which are not used by any family parameter in the family or nested families.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param processor: An RevitSharedParameterDataProcessor object containing all shared parameter information of the family document and any nested families.
    :type processor: :class:`.SharedParameterProcessor`

    :return:
        Result class instance.

        - True if all unused shared parameters where deleted successfully or none needed to be deleted. Otherwise False.
        - Result.message property updated in format: Found unused shared parameter: shared parameter Name [GUID]

        On exception:

        - status (bool) will be False.
        - message will contain the exception message.

    :rtype: :class:`.Result`
    """

    # from processor instance get all root line pattern entries where usage counter == 0.
    # delete those line patterns by id

    return_value = res.Result()

    # check the category of the family first:
    # Tags and generic annotations (may) contain labels which in turn use shared parameters to drive them
    # there is currently no way in the api (Revit 2022) to find out what parameter is driving the label...

    # get the family category name:
    fam_cat_name = list(rCat.get_family_category(doc))[0]
    if fam_cat_name != "Generic Annotations" and fam_cat_name.endswith("Tags") == False:
        ids_to_delete = []
        # get categories found in root processor data only
        root_fam_data = processor._find_root_family_data()
        # get all root line pattern entries where usage counter == 0.
        for root_fam in root_fam_data:
            if root_fam[IFamData.USAGE_COUNTER] == 0:
                return_value.append_message(
                    "Found unused shared parameter: {} [{}]".format(
                        root_fam[rSharedParaData.PARAMETER_NAME],
                        root_fam[rSharedParaData.PARAMETER_GUID],
                    )
                )
                ids_to_delete.append(
                    rdb.ElementId(root_fam[rSharedParaData.PARAMETER_ID])
                )
        # delete any subcategories found
        if len(ids_to_delete) > 0:
            result_delete = rDel.delete_by_element_ids(
                doc,
                ids_to_delete,
                "Deleting unused shared parameters.",
                "Shared Parameters",
            )
            return_value.update(result_delete)
            # may need to delete shared parameters one by one if one or more cant be deleted
            if result_delete.status == False:
                result_delete_one_by_one = rDel.delete_by_element_ids_one_by_one(
                    doc,
                    ids_to_delete,
                    "Deleting unused shared parameters: one by one.",
                    "Shared Parameters",
                )
                return_value.update(result_delete_one_by_one)
        else:
            return_value.update_sep(
                True, "No unused shared parameters found. Nothing was deleted."
            )
    else:
        return_value.update_sep(
            True,
            "This is an annotation family (tag or generic annotation). Due to limitations in the Revit API no shared parameter was purged.",
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit groups helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)
import System

# import common library modules
from duHast.Revit.Common import common as com

# import Autodesk
from Autodesk.Revit.DB import BuiltInCategory, FilteredElementCollector

# --------------------------------------------- utility functions ------------------


# doc   current document
def get_model_groups(doc):
    """
    Get all model group types from the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: list of model group types in the model
    :rtype: list of Autodesk.Revit.DB.
    """

    return (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_IOSModelGroups)
        .WhereElementIsElementType()
        .ToList()
    )


def get_model_group_instances(doc):
    """
    Get all model group instances in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: collector of model group instances
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """
    col = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_IOSModelGroups)
        .WhereElementIsNotElementType()
    )
    return col


def get_model_group_instances_by_type(doc):
    """
    Get all model group instances in the model grouped by their type id.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: Dictionary of model group instances by type id
    :rtype: {ElementId:[Autodesk.Revit.DB.Group]}
    """

    model_groups_by_type_id = {}
    model_group_instances = get_model_group_instances(doc=doc)
    for instance in model_group_instances:
        type_id = instance.GetTypeId()
        if type_id in model_groups_by_type_id:
            model_groups_by_type_id[type_id].append(instance)
        else:
            model_groups_by_type_id[type_id] = [instance]
    return model_groups_by_type_id


def get_detail_groups(doc):
    """
    Gets all detail groups in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list containing detail group types.
    :rtype: list
    """

    return (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_IOSDetailGroups)
        .WhereElementIsElementType()
        .ToList()
    )


def get_nested_detail_groups(doc):
    """
    Gets all nested detail groups in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A list containing nested detail group types.
    :rtype: list
    """

    return (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_IOSAttachedDetailGroups)
        .WhereElementIsElementType()
        .ToList()
    )


def get_model_group_ids(doc):
    """
    Gets a list of all model group type ids in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of model group type ids
    :rtype: List of Autodesk.Revit.DB.ElementId
    """

    ids = []
    col = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_IOSModelGroups)
        .WhereElementIsElementType()
    )
    ids = com.get_ids_from_element_collector(col)
    return ids


def get_detail_group_ids(doc):
    """
    Gets a list of all detail group types from the model.

    This will not include any attached detail groups.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of detail group type ids
    :rtype: List of Autodesk.Revit.DB.ElementId
    """

    ids = []
    col = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_IOSDetailGroups)
        .WhereElementIsElementType()
    )
    ids = com.get_ids_from_element_collector(col)
    return ids


def get_nested_detail_group_ids(doc):
    """
    Gets a list of all nested detail group types from the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of nested detail group type ids
    :rtype: List of Autodesk.Revit.DB.ElementId
    """

    ids = []
    col = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_IOSAttachedDetailGroups)
        .WhereElementIsElementType()
    )
    ids = com.get_ids_from_element_collector(col)
    return ids


def get_unplaced_groups(doc, group_category):
    """
    Gets a list of unplaced groups from the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param group_category: A built in category defining the group category (model vs detail)
    :type group_category: either BuiltInCategory.OST_IOSDetailGroups or BuiltInCategory.OST_IOSModelGroups

    :return: List of unplaced group types
    :rtype: list
    """

    def getter_types(doc):
        return (
            FilteredElementCollector(doc)
            .OfCategory(group_category)
            .WhereElementIsElementType()
        )

    def getter_instances(doc):
        return (
            FilteredElementCollector(doc)
            .OfCategory(group_category)
            .WhereElementIsNotElementType()
        )

    # get unplaced groups
    return com.get_not_placed_types(doc, getter_types, getter_instances)


def get_unplaced_detail_groups(doc):
    """
    Gets a list of unplaced detail groups from the model

    This will not include any attached detail groups.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of unplaced detail group types
    :rtype: list
    """

    return get_unplaced_groups(doc, BuiltInCategory.OST_IOSDetailGroups)


def get_unplaced_detail_group_ids(doc):
    """
    Gets a list of unplaced detail groups type Ids from the model.

    This will not include any attached detail groups.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of unplaced detail group type ids
    :rtype: List of Autodesk.Revit.DB.ElementId
    """

    unplaced_groups = get_unplaced_groups(doc, BuiltInCategory.OST_IOSDetailGroups)
    ids = []
    for unplaced in unplaced_groups:
        ids.append(unplaced.Id)
    return ids


def get_unplaced_nested_detail_groups(doc):
    """
    Gets a list of unplaced nested detail groups from the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of unplaced nested detail group types
    :rtype: list
    """

    return get_unplaced_groups(doc, BuiltInCategory.OST_IOSAttachedDetailGroups)


def get_unplaced_nested_detail_group_ids(doc):
    """
    Gets a list of unplaced nested detail group Ids from the model.

    This will not list any none nested detail groups.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of unplaced model group type ids
    :rtype: List of Autodesk.Revit.DB.ElementId
    """

    unplaced_groups = get_unplaced_groups(
        doc, BuiltInCategory.OST_IOSAttachedDetailGroups
    )
    ids = []
    for unplaced in unplaced_groups:
        ids.append(unplaced.Id)
    return ids


def get_unplaced_model_groups(doc):
    """
    Gets a list of unplaced model groups types from the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of unplaced model group types
    :rtype: list
    """

    return get_unplaced_groups(doc, BuiltInCategory.OST_IOSModelGroups)


def get_unplaced_model_group_ids(doc):
    """
    Gets a list of unplaced model group type Ids from the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of unplaced model group type ids
    :rtype: List of Autodesk.Revit.DB.ElementId
    """

    unplaced_groups = get_unplaced_groups(doc, BuiltInCategory.OST_IOSModelGroups)
    ids = []
    for unplaced in unplaced_groups:
        ids.append(unplaced.Id)
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a Revit tag instances report function. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import json

from duHast.Revit.Annotation.independent_tags import get_all_independent_tags
from duHast.Revit.Annotation.independent_tags_elbow_properties import (
    get_elbow_properties,
)
from duHast.Revit.Annotation.independent_tags_tagged_elements import get_tagged_elements

from duHast.Revit.Annotation.Reporting import (
    gen_annotations_instance_report_header as props,
)

from duHast.Revit.Common.Geometry.points import get_point_as_doubles
from duHast.Revit.Common.revit_version import get_revit_version_number

import Autodesk.Revit.DB as rdb


def _convert_tagged_element_ids_to_int(data):
    """
    Converts list of dictionaries containing element ids as values to integer values so it can be converted to json

    :param data: A list of dictionaries where values are Element ids
    :type data: [{str:Autodesk.Revit.DB.ElementId}]
    :return: A list of dictionaries where values are int
    :rtype: [{str:int}]
    """

    for tag_dic in data:
        for id_entry in tag_dic:
            if type(tag_dic[id_entry]) == rdb.ElementId:
                tag_dic[id_entry] = tag_dic[id_entry].IntegerValue
    return data


def get_tag_instances_report_data(doc, revit_file_path, custom_element_filter):
    """
    Gets tag instances data to be written to report file.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_file_path: The file hostname, which is added to data returned.
    :type revit_file_path: str
    :param custom_element_filter: allows to filter tags by specific tag properties
    :type custom_element_filter: class:`.RevitCustomElementFilter`

    :return: list of list of sheet properties.
    :rtype: list of list of str
    """

    data = []
    revit_version = get_revit_version_number(doc)
    tag_instances = get_all_independent_tags(doc)
    for tag_instance in tag_instances:
        row = {}
        try:
            if custom_element_filter != None:
                if custom_element_filter.check_element(doc, tag_instance.Id):
                    # this can throw an exception...wrap in try catch
                    tag_text = "<empty>"
                    try:
                        tag_text = tag_instance.TagText
                    except:
                        pass

                    # get tagged element ids
                    tagged_element_ids = get_tagged_elements(doc=doc, tag=tag_instance)
                    # convert ids to integer value ( to be able to output them to json)
                    tagged_element_data = _convert_tagged_element_ids_to_int(
                        tagged_element_ids
                    )
                    # get elbow properties
                    elbow_properties = None
                    # get elbow properties
                    if tag_instance.HasLeader:
                        elbow_properties = get_elbow_properties(
                            doc=doc, tag=tag_instance, points_as_double=True
                        )

                    # leader end condition (need to check if there is a leader)
                    leader_end_condition = str(None)
                    if tag_instance.HasLeader:
                        leader_end_condition = str(tag_instance.LeaderEndCondition)

                    # base line revit data
                    row = {
                        props.HOST_FILE: revit_file_path,
                        props.TAG_ID: tag_instance.Id.IntegerValue,
                        props.TAG_HAS_LEADER: tag_instance.HasLeader,  # leader flag
                        props.TAG_IS_ORPHANED: tag_instance.IsOrphaned,  # is orphaned tag?
                        props.TAG_IS_MATERIAL_TAG: tag_instance.IsMaterialTag,  # is a material tag
                        props.IS_MULTICATEGORY_TAG: tag_instance.IsMulticategoryTag,  # is is multi category tag
                        props.LEADER_END_CONDITION: leader_end_condition,  # attached or free
                        props.LEADER_PROPERTIES: elbow_properties,  # elbow properties
                        props.MULTI_REFERENCE_ANNOTATION_ID: tag_instance.MultiReferenceAnnotationId.IntegerValue,
                        props.TAG_TEXT: tag_text,  # tag text
                        props.TAGGED_ELEMENT_IDS: tagged_element_data,  # tagged element ids as integers
                        props.TAG_HEAD_LOCATION: get_point_as_doubles(
                            tag_instance.TagHeadPosition
                        ),  # tag location
                        props.TAG_ROTATION_ANGLE: tag_instance.RotationAngle,  # rotation tag
                        props.TAG_ORIENTATION: str(
                            tag_instance.TagOrientation
                        ),  # horizontal ,vertical, model
                    }

                    # add data for later versions of revit
                    if revit_version >= 2023:
                        row[props.TAG_PRESENTATION_MODE] = str(
                            tag_instance.LeadersPresentationMode
                        )
                        row[props.MERGE_ELBOWS] = tag_instance.MergeElbows
                    data.append(row)
        except Exception as e:
            data.append(
                {
                    props.HOST_FILE: revit_file_path,
                    props.TAG_ID: str(tag_instance.Id),
                    "exception": str(e),
                }
            )
    return data


def read_tag_independent_data_from_file(revit_file_path):
    """
    Reads an independent tags report file into a list of dictionaries

    :param revit_file_path: Fully qualified file path of report file.
    :type revit_file_path: str
    :return: List of dictionaries where each dictionary contains tag data
    :rtype: [{}]
    """

    data = {}
    try:
        # Opening JSON file
        f = open(revit_file_path)
        # returns JSON object as
        # a dictionary
        data = json.load(f)
    except Exception as e:
        pass
    return data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit fill patterns helper functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import ElementId, FilteredElementCollector, FillPatternElement

from duHast.Revit.DetailItems.detail_items import get_all_filled_region_type_ids_available

def get_all_fill_pattern(doc):
    """
    Gets all fill pattern elements in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector of all fill pattern elements.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    return FilteredElementCollector(doc).OfClass(FillPatternElement)


def pattern_ids_by_name(doc):
    """
    Returns a dictionary where fill pattern name is key, values are all ids of line patterns with the exact same name.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A dictionary where fill pattern name is key, values are all ids of line patterns with the exact same name
    :rtype: dictionary(key str, value list of Autodesk.Revit.DB.ElementId)
    """

    pattern_dic = {}
    all_fill_pattern = get_all_fill_pattern(doc=doc)
    for fill_pattern in all_fill_pattern:
        pattern_name = fill_pattern.GetFillPattern().Name
        if pattern_name in pattern_dic:
            pattern_dic[pattern_name].append(fill_pattern.Id)
        else:
            pattern_dic[pattern_name] = [fill_pattern.Id]
    return pattern_dic

def get_used_pattern_ids_in_filled_region_types(doc):
    """
    Returns a list of all fill pattern ids used in filled region types in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list of all fill pattern ids used in filled region types in the model.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """
    
    fill_pattern_ids = []
    # get all filled region types in model
    filled_region_type_ids =  get_all_filled_region_type_ids_available(doc)
    
    # return an empty list of no filled region types are available
    if(len(filled_region_type_ids)==0):
        return fill_pattern_ids
        
    for filled_region_type_id in filled_region_type_ids:
        # get the filled region type and its fill pattern ids for fore and background
        filled_region_type = doc.GetElement(filled_region_type_id)
        fore_ground_pattern_id = filled_region_type.ForegroundPatternId
        background_pattern_id = filled_region_type.BackgroundPatternId
        # check if already in unique list of fill pattern ids
        if(fore_ground_pattern_id not in fill_pattern_ids and fore_ground_pattern_id != ElementId.InvalidElementId):
            fill_pattern_ids.append(fore_ground_pattern_id)
        if(background_pattern_id not in fill_pattern_ids and background_pattern_id != ElementId.InvalidElementId):
            fill_pattern_ids.append(background_pattern_id)

    return fill_pattern_ids


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions relating to directory IO . 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import os
from os import path
import os.path
import shutil


def directory_empty_delete(full_directory_path):
    """
    Deletes an empty directory
    :param full_directory_path: Path to directory
    :type full_directory_path: str
    :return: True directory deleted, otherwise False
    :rtype: bool
    """

    try:
        os.rmdir(full_directory_path)
        value = True
    except Exception:
        value = False
    return value


def directory_delete(full_directory_path):
    """
    Deletes a directory (even if it contains files)
    :param full_directory_path: Path to directory
    :type full_directory_path: str
    :return: True directory deleted, otherwise False
    :rtype: bool
    """

    try:
        shutil.rmtree(full_directory_path)
        value = True
    except Exception as e:
        print(
            "When attempting to delete: {} an exception occurred: {}".format(
                full_directory_path, e
            )
        )
        value = False
    return value


def get_child_directories(full_directory_path):
    """
    Returns the immediate subdirectories of directory
    :param full_directory_path: Path to directory
    :type full_directory_path: str
    :return: any sub directories, empty list if none exist
    :rtype: list of str
    """

    sub_folders_with_paths = []
    for root, dirs, files in os.walk(full_directory_path):
        for dir in dirs:
            sub_folders_with_paths.append(os.path.join(root, dir))
        break
    return sub_folders_with_paths


def get_parent_directory(full_directory_path):
    """
    Returns the parent directory of directory, or empty string if invalid directory
    :param full_directory_path: Path to directory
    :type full_directory_path: str
    :return: parent directory, or empty string
    :rtype: str
    """

    parent_dir = ""
    try:
        parent_dir = os.path.dirname(full_directory_path)
    except Exception:
        pass
    return parent_dir


def create_directory(root, folder_name):
    """
    Create a folder.
    :param root: Directory path in which the new folder is to be created
    :type root: str
    :param folder_name: New folder name.
    :type folder_name: str
    :return: True if folder is created, otherwise False
    :rtype: bool
    """

    dir_name = path.join(root, folder_name)
    flag = True
    try:
        os.mkdir(dir_name)
    except Exception as e:
        # just in case the folder does exist (created by another instance at almost the same time)
        if "[Errno 17]" not in str(e):
            flag = False
    return flag


def create_target_directory(root_path, folder_name):
    """
    Create a folder.
    Checks whether folder exists and if not attempts to create it.

    :param root: Directory path in which the new folder is to be created
    :type root: str
    :param folder_name: New folder name.
    :type folder_name: str
    :return: True if folder is created, otherwise False
    :rtype: bool
    """

    # check if folder exists
    flag = True
    if path.exists(root_path + "\\" + folder_name) == False:
        # create new folder
        flag = create_directory(root_path, folder_name)
    return flag


def directory_exists(directory_path):
    """
    Check if a given directory exists
    :param directory_path: Fully qualified directory path
    :type directory_path: str
    :return: True if directory exists, otherwise False
    :rtype: bool
    """
    if path.exists(directory_path):
        return True
    else:
        return False



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family Base data analysis module containing functions to find missing families.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

A missing family is a family which is not present in the family base data report as a root family but present as a nested family.

Algorithm description:

- read base data file processing list (created by RevitFamilyBaseDataProcessor module)
- extract all root families (no :: in root path)
- extract all nested families ( :: in root path)

- cull nested family so each nested family appears only once

- loop over nested families
    - loop over root families 
        - check if nested family exists in root families ( by name and category)
            - no
            - add to missing families list
            - stop processing since one match is enough per missing family
- write out missing families data ( host family filepath,)

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# from collections import namedtuple

from duHast.Revit.Family.Data import family_base_data_utils as rFamBaseDataUtils
from duHast.Utilities.Objects.timer import Timer
from duHast.Utilities.Objects import result as res


def _build_unique_nested_family_dic(overallFamilyBaseNestedData):
    """
    Builds a dictionary containing unique nested families only. Key is the family name and category concatenated.

    :param overallFamilyBaseNestedData: List of tuples containing nested family data.
    :type overallFamilyBaseNestedData: [nestedFamily]

    :return: A dictionary where:

        - Key is the family name and category concatenated.
        - value is a tuple representing one instance of the nested family data.
    :rtype:{ str : nestedFamily }
    """

    uniqueDic = {}

    for nested in overallFamilyBaseNestedData:
        keyNew = nested.name + nested.category
        if keyNew not in uniqueDic:
            uniqueDic[keyNew] = nested

    return uniqueDic


def _check_nested_against_root_families(
    uniqNestedDictionary, overallFamilyBaseRootData
):
    """
    Compared a dictionary containing tuples containing nested family data against list of root family data.
    Any nested family without a match in the root family data list will be returned (missing from list).

    :param uniqNestedDictionary: A dictionary where:

        - Key is the family name and category concatenated.
        - value is a tuple representing one instance of the nested family data.

    :type uniqNestedDictionary: { str : nestedFamily }
    :param overallFamilyBaseRootData: List of tuples containing root family data.
    :type overallFamilyBaseRootData: [rootFamily]

    :return: List of tuples containing nested family data.
    :rtype: [nestedFamily]
    """

    missingFamilies = []
    for nestedId, nestedFam in uniqNestedDictionary.items():
        match = False
        for baseRootFam in overallFamilyBaseRootData:
            if (
                nestedFam.name == baseRootFam.name
                and nestedFam.category == baseRootFam.category
            ):
                match = True
                break
        if match == False:
            missingFamilies.append(nestedFam)
    return missingFamilies


def _get_unique_nested_families(overallFamilyBaseNestedData):
    """
    Loops over family nested base data and creates a unique list of nested families based on name and category.
    This is done to make search faster (reduce the to be searched data set)

    :param overallFamilyBaseNestedData: List of tuples containing nested family data.
    :type overallFamilyBaseNestedData: [nestedFamily]
    :return: Unique list of tuples containing nested family data.
    :rtype: [nestedFamily]
    """

    uniqueNestedFamiliesCompare = []
    uniqueNestedFamilies = []
    for nestedFam in overallFamilyBaseNestedData:
        if nestedFam.name + nestedFam.category not in uniqueNestedFamiliesCompare:
            uniqueNestedFamilies.append(nestedFam)
            uniqueNestedFamiliesCompare.append(nestedFam.name + nestedFam.category)

    return uniqueNestedFamilies


def check_families_missing_from_library(familyBaseDataReportFilePath):
    """
    Processes a family base data report and identifies any nested families which have not been processed as a root family\
        and therefore do not exist in the library.
    
    :param familyBaseDataReportFilePath: Fully qualified file path to family base data report file. 
    :type familyBaseDataReportFilePath: str
    
    :return: 
        Result class instance.

        - result.status. True if missing families where found without an exception occurring.
        - result.message will contain the summary messages of the process including time stamps.
        - result.result [nestedFamily]
        
        On exception:
        
        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.result will be empty

    :rtype: :class:`.Result`
    """
    returnValue = res.Result()
    try:
        # set up a timer
        tProcess = Timer()
        tProcess.start()

        # read overall family base data from file
        (
            overallFamilyBaseRootData,
            overallFamilyBaseNestedData,
        ) = rFamBaseDataUtils.read_overall_family_data_list(
            familyBaseDataReportFilePath
        )
        returnValue.append_message(
            "{} Read overall family base data report. {} root entries found and {} nested entries found.".format(
                tProcess.stop(),
                len(overallFamilyBaseRootData),
                len(overallFamilyBaseNestedData),
            )
        )

        tProcess.start()
        # get a list of unique families from the nested family data
        uniqueFamilyBaseNestedData = _get_unique_nested_families(
            overallFamilyBaseNestedData
        )
        returnValue.append_message(
            "{} Culled nested family base data from : {} to: {} families".format(
                tProcess.stop(),
                len(overallFamilyBaseNestedData),
                len(uniqueFamilyBaseNestedData),
            )
        )

        tProcess.start()
        # read over nested data and built a list of unique families ( name + category )
        uniqueDic = _build_unique_nested_family_dic(uniqueFamilyBaseNestedData)
        returnValue.append_message(
            "{} Found unique nested families [{}]".format(
                tProcess.stop(), len(uniqueDic)
            )
        )

        tProcess.start()
        # identify missing families in unique list of nested families
        missingFamilies = _check_nested_against_root_families(
            uniqueDic, overallFamilyBaseRootData
        )
        returnValue.append_message(
            "{} Found: {} missing families.".format(
                tProcess.stop(), len(missingFamilies)
            )
        )
        if len(missingFamilies) > 0:
            returnValue.result = missingFamilies
    except Exception as e:
        returnValue.update_sep(
            False, "Failed to retrieve missing families with exception: {}".format(e)
        )
    return returnValue


# ----------------------------missing families: direct host files -----------------------------------------


def find_missing_families_direct_host_families(
    familyBaseDataReportFilePath, missingFamilies
):
    """
    Returns a list of root family tuples which represent the direct parents (host families) of the missing families.

    :param missingFamilies: A list of tuple containing nested family data representing missing families(no base root family entry)
    :type missingFamilies: [nestedFamily]
    :param familyBaseDataReportFilePath: Fully qualified file path to family base data report file.
    :type familyBaseDataReportFilePath: str

    :return:
        Result class instance.

        - result.status. True if host families of missing families where found without an exception occurring.
        - result.message will contain the summary messages of the process including time stamps.
        - result.result [rootFamily]

        On exception:

        - result.status (bool) will be False.
        - result.message will contain generic exception message.
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    # loop over missing families
    # loop over base nested data
    #   - check if missing fam is in root path (name and category) if so:
    #   - get the direct parent (make sure missing family isn't first entry!)
    #   - check if direct parent is already in dictionary (key is name + category) ? if not:
    #       - add direct parent to dictionary
    #
    # loop over direct host data:
    #   - loop over root fam data and check for match in name and category; If so:
    #       - add to root family data to be returned.

    returnValue = res.Result()
    try:
        # set up a timer
        tProcess = Timer()
        tProcess.start()

        returnValue = res.Result()
        # read overall family base data from file
        (
            overallFamilyBaseRootData,
            overallFamilyBaseNestedData,
        ) = rFamBaseDataUtils.read_overall_family_data_list(
            familyBaseDataReportFilePath
        )
        returnValue.append_message(
            "{} Read overall family base data report. {} root entries found and {} nested entries found.".format(
                tProcess.stop(),
                len(overallFamilyBaseRootData),
                len(overallFamilyBaseNestedData),
            )
        )

        tProcess.start()
        hostFamilies = rFamBaseDataUtils.find_all_direct_host_families(
            missingFamilies, overallFamilyBaseNestedData
        )
        # get the root families from host family data
        rootHosts = rFamBaseDataUtils.find_root_families_from_hosts(
            hostFamilies, overallFamilyBaseRootData
        )
        returnValue.result = rootHosts
        returnValue.append_message(
            "{} Found direct host families of missing families: {}".format(
                tProcess.stop(), len(rootHosts)
            )
        )
    except Exception as e:
        returnValue.update_sep(
            False,
            "Failed to retrieve host families of missing families with exception: ".format(
                e
            ),
        )
    return returnValue



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions relating to Solibri IFC optimizer.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

List of imports:

- :class:`.Result`
- :module: Utility

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import subprocess
from System.IO import Path

from duHast.Utilities.Objects import result as res
from duHast.Utilities import files_get as fileGet, files_io as fileIO


#: The default install path for solibri ifc optimizer.
solibri_install_path_ = (
    r"C:\Program Files\Solibri\IFCOptimizer\Solibri IFC Optimizer.exe"
)


def optimize_all_ifc_files_in_directory(directory_path):
    """
    Function applying third party IFC optimizer to all ifc files in a given folder.

    Original files will be deleted.

    :param directory_path: The directory path where IFC files are located
    :type directory_path: str
    :return:
        Result class instance.

        - Optimizer status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the fully qualified file path(s) of the optimized file(s).

        On exception (handled by optimizer itself!):

        - result.status (bool) will be False.
        - result.message will contain generic exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # check if ifc optimizer is installed:
    if fileIO.file_exist(solibri_install_path_):
        return_value.message = "Solibri IFC optimizer is installed."
        ifc_files = fileGet.get_files(directory_path, ".ifc")
        if len(ifc_files) > 0:
            process_files_result = process_ifc_files(ifc_files, directory_path)
            return_value.update(process_files_result)
        else:
            return_value.append_message(
                "No IFC files found in directory: {}".format(directory_path)
            )
    else:
        return_value.update_sep(
            False, "No IFC optimizer installed at: {}".format(solibri_install_path_)
        )
    return return_value


def optimize_ifc_files_in_list(ifc_files, directory_path):
    """
    This function will optimize all IFC files in a given list of fully qualified file path to ifc files.

    Will check whether Solibri IFC optimizer is installed.

    :param ifc_files: List containing fully qualified file path of ifc files to be optimized.
    :type ifc_files: list of str
    :param directory_path: Directory of where the optimized IFC file(s) are to be saved.
    :type directory_path: str

    :return:
        Result class instance.

        - Optimizer status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the fully qualified file path(s) of the optimized file(s).

        On exception (handled by optimizer itself!):

        - result.status (bool) will be False.
        - result.message will contain generic exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # check if ifc optimizer is installed:
    if fileIO.file_exist(solibri_install_path_):
        return_value.message = "Solibri IFC optimizer is installed."
        if len(ifc_files) > 0:
            process_files_result = process_ifc_files(ifc_files, directory_path)
            return_value.update(process_files_result)
        else:
            return_value.append_message("IFC file list is empty.")
    else:
        return_value.update_sep(
            False, "No IFC optimizer installed at: " + str(solibri_install_path_)
        )
    return return_value


def process_ifc_files(ifc_files, directory_path):
    """
    This function will optimize all IFC files in a given list of fully qualified file path to ifc files.

    Will not check whether Solibri IFC optimizer is installed.

    :param ifc_files: List containing fully qualified file path of ifc files to be optimized.
    :type ifc_files: list of str
    :param directory_path: Directory of where the optimized IFC file(s) are to be saved.
    :type directory_path: str

    :return:
        Result class instance.

        - Optimizer status returned in result.status. False if an exception occurred, otherwise True.
        - result.message will contain the fully qualified file path(s) of the optimized file(s).

        On exception (handled by optimizer itself!):

        - result.status (bool) will be False.
        - result.message will contain generic exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    files_to_delete = []
    files_to_rename = []
    if len(ifc_files) > 0:
        return_value.append_message("found ifc files: {}".format(len(ifc_files)))
        for ifc_file in ifc_files:
            s = subprocess.check_call(
                [
                    r"C:\Program Files\Solibri\IFCOptimizer\Solibri IFC Optimizer.exe",
                    "-in=" + ifc_file,
                    "-out=" + directory_path,
                    "-ifc",
                    "-force",
                ]
            )
            # check what came back
            if s == 0:
                # all went ok:
                return_value.append_message("Optimized file: {}".format(ifc_file))
                files_to_delete.append(ifc_file)  # full file path
                # get the rename information
                # contains old and new file name
                rename = []
                p = fileIO.get_directory_path_from_file_path(ifc_file)
                if p != "":
                    new_file_path = (
                        str(p)
                        + "\\"
                        + str(Path.GetFileNameWithoutExtension(ifc_file))
                        + "_optimized.ifc"
                    )
                    rename.append(new_file_path)
                    rename.append(ifc_file)
                    files_to_rename.append(rename)
            else:
                # something went wrong
                return_value.update_sep(
                    False, "Failed to optimize file: {}".format(ifc_file)
                )
        # clean up
        for file_to_delete in files_to_delete:
            status_delete = fileIO.file_delete(file_to_delete)
            if status_delete:
                return_value.append_message(
                    "Deleted original file: {}".format(file_to_delete)
                )
            else:
                return_value.update_sep(
                    False, "Failed to delete original file: {}".format(file_to_delete)
                )
        for file_to_rename in files_to_rename:
            status_rename = fileIO.rename_file(file_to_rename[0], file_to_rename[1])
            if status_rename:
                return_value.append_message(
                    "Renamed original file: {} to: {}".format(
                        file_to_rename[0], file_to_rename[1]
                    )
                )
            else:
                return_value.update_sep(
                    False,
                    "Failed to rename original file: {}".format(file_to_rename[0]),
                )
    else:
        return_value.append_message("No IFC files found at: {}".format(directory_path))
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit elements to category helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Revit.Common import parameter_set_utils as rParaSet
from duHast.Utilities.Objects import result as res
from duHast.Revit.Links import cad_links_geometry as rCadLinkGeo
from duHast.Revit.Common import parameter_get_utils as rParaGet

from duHast.Revit.Categories.categories import (
    ELEMENTS_PARAS_SUB,
    get_main_sub_categories,
)
from duHast.Revit.Categories.Utility.category_properties_get_utils import (
    get_category_graphic_style_ids,
)
from duHast.Revit.Categories.Utility.category_property_names import (
    CATEGORY_GRAPHIC_STYLE_3D,
)
from duHast.Revit.Family import family_element_utils as rFamElementUtils


def sort_elements_by_category(elements, element_dic):
    """
    Returns a dictionary of element ids where key is the category they belong to.
    :param elements:  List of revit elements.
    :type elements: [Autodesk.Revit.DB.Element]
    :param element_dic:  Dictionary where key is subcategory and values are element ids.
    :type element_dic: {Autodesk.Revit.DB.Category: [Autodesk.Revit.DB.ElementId]}
    :return: Dictionary where key is subcategory id and values are element ids.
    :rtype: {Autodesk.Revit.DB.ElementId: [Autodesk.Revit.DB.ElementId]}
    """

    for el in elements:
        for builtin_def in ELEMENTS_PARAS_SUB:
            value = rParaGet.get_built_in_parameter_value(
                el, builtin_def, rParaGet.get_parameter_value_as_element_id
            )
            if value != None:
                if value in element_dic:
                    element_dic[value].append(el.Id)
                else:
                    element_dic[value] = [el.Id]
                break
    return element_dic


def sort_geometry_elements_by_category(elements, element_dic, doc):
    """
    Sorts geometry elements by their category in a Revit family document.

    Args:
        elements (list): A list of geometry elements to be sorted.
        element_dic (dict): A dictionary where the sorted elements will be stored.
        doc (Revit Document): The current Revit family document.

    Returns:
        dict: A dictionary where the keys are category ids and the values are lists of element ids. The geometry elements are sorted by their category.
    """
    
    counter = 0
    for el in elements:
        counter = counter + 1
        graphic_style_id = rdb.ElementId.InvalidElementId
        if type(el) is rdb.Solid:
            # get graphic style id from edges
            edge_array = el.Edges
            if edge_array.IsEmpty == False:
                for edge in edge_array:
                    graphic_style_id = edge.GraphicsStyleId
        else:
            graphic_style_id = el.GraphicsStyleId
        # failed to get an id?
        if graphic_style_id != rdb.ElementId.InvalidElementId:
            graphic_style = doc.GetElement(graphic_style_id)
            graph_cat_id = graphic_style.GraphicsStyleCategory.Id
            # geometry elements have no Id property ... Doh!! pass in invalid element id...
            if graph_cat_id != None:
                if graph_cat_id in element_dic:
                    element_dic[graph_cat_id].append(rdb.ElementId.InvalidElementId)
                else:
                    element_dic[graph_cat_id] = [rdb.ElementId.InvalidElementId]
    return element_dic


def _sort_all_elements_by_category(doc):
    """
    Sorts all elements in a family by category.
    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :return: Dictionary where key is subcategory id and values are element ids.
    :rtype: {Autodesk.Revit.DB.ElementId: [Autodesk.Revit.DB.ElementId]}
    """

    # get all elements in family
    dic = {}
    el_curve = rFamElementUtils.get_all_curve_based_elements_in_family(doc)
    el_forms = rFamElementUtils.get_all_generic_forms_in_family(doc)
    el_m_text = rFamElementUtils.get_all_model_text_elements_in_family(doc)
    el_ref_planes = rFamElementUtils.get_all_reference_planes_in_family(doc)
    # get import Instance elements
    el_import = rCadLinkGeo.get_all_cad_import_instances_geometry(doc)
    # build dictionary where key is category or graphic style id of  a category
    dic = sort_elements_by_category(el_curve, dic)
    dic = sort_elements_by_category(el_forms, dic)
    dic = sort_elements_by_category(el_m_text, dic)
    dic = sort_elements_by_category(el_ref_planes, dic)
    # geometry instances use a property rather then a parameter to store the category style Id
    dic = sort_geometry_elements_by_category(el_import, dic, doc)
    return dic


def get_elements_by_category(doc, cat):
    """
    Returns elements in family assigned to a specific category
    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param cat: A category.
    :type cat: Autodesk.Revit.DB.Category
    :return: Dictionary where key is subcategory and values are element ids.
    :rtype: {Autodesk.Revit.DB.Category: [Autodesk.Revit.DB.ElementId]}
    """

    # get all elements in family
    dic = _sort_all_elements_by_category(doc)
    # get id and graphic style id of category to be filtered by
    category_ids = get_category_graphic_style_ids(cat)
    # check whether category past in is same as owner family category
    if doc.OwnerFamily.FamilyCategory.Name == cat.Name:
        # 3d elements within family which have subcategory set to 'none' belong to owner family
        # category. Revit uses a None value as id rather then the actual category id
        # my get parameter value translates that into -1 (invalid element id)
        category_ids[CATEGORY_GRAPHIC_STYLE_3D] = rdb.ElementId.InvalidElementId
    dic_filtered = {}
    # filter elements by category ids
    for key, value in category_ids.items():
        # print (key + ' ' + str(value))
        if value in dic:
            dic_filtered[key] = dic[value]
        else:
            dic_filtered[key] = []
    return dic_filtered


def move_elements_to_category(doc, elements, to_category_name, destination_cat_ids):
    """
    Moves elements provided in a dictionary to another category specified by name.

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param elements: Dictionary of elements, where keys are graphic style names.
    :type elements: {Autodesk.Revit.DB.Category: [Autodesk.Revit.DB.ElementId]}
    :param to_category_name: The name of the subcategory elements are to be moved to.
    :type to_category_name: str
    :param destination_cat_ids: Dictionary of IDs of graphic styles, where keys are graphic style names.
    :type destination_cat_ids: dictionary {str: Autodesk.Revit.DB.ElementId}

    :return: Result class instance.
        - `result.status` (bool): True if all elements were moved to destination subcategories, otherwise False.
        - `result.message` (str): Contains the name of the destination subcategory by element.
        - `result.result` (empty list)

    On exception:
        - `result.status` (bool): False.
        - `result.message` (str): Contains a generic exception message.
        - `result.result` (empty list)

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # check whether destination category exist in file
    cats = get_main_sub_categories(doc)
    if to_category_name in cats:
        for key, value in elements.items():
            # anything needing moving?
            if len(value) > 0:
                for el_id in value:
                    el = doc.GetElement(el_id)
                    paras = el.GetOrderedParameters()
                    # find the parameter driving the subcategory
                    for p in paras:
                        if p.Definition.BuiltInParameter in ELEMENTS_PARAS_SUB:
                            # get the subcategory style id
                            target_id = destination_cat_ids[key]
                            # check if a 'cut' style id exists...if not move to 'projection' instead
                            # not sure how this works in none - english versions of Revit...
                            if (
                                key == "Cut"
                                and target_id == rdb.ElementId.InvalidElementId
                            ):
                                target_id = destination_cat_ids["Projection"]
                                return_value.append_message(
                                    "No cut style present in family, using projection style instead"
                                )
                            updated_para = rParaSet.set_parameter_value(
                                p, str(target_id), doc
                            )
                            return_value.update(updated_para)
                            break
    else:
        return_value.update_sep(
            False,
            "Destination category: {} does not exist in file!".format(to_category_name),
        )
    return return_value


def move_elements_from_sub_category_to_sub_category(
    doc, from_category_name, to_category_name
):
    """
    Moves elements from one subcategory to another one identified by their names.

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param from_category_name: The source subcategory name.
    :type from_category_name: str
    :param to_category_name: The destination subcategory name.
    :type to_category_name: str

    :return: Result class instance.
        - `result.status` (bool): True if all elements from the source subcategory were moved to the destination subcategory, otherwise False.
        - `result.message` (str): Contains the name of the destination subcategory by element.
        - `result.result` (empty list)

    On exception:
        - `result.status` (bool): False.
        - `result.message` (str): Contains a generic exception message.
        - `result.result` (empty list)

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # check whether source and destination category exist in file
    cats = get_main_sub_categories(doc)
    if from_category_name in cats:
        if to_category_name in cats:
            # dictionary containing destination category ids (3D, cut and projection)
            destination_cat_ids = get_category_graphic_style_ids(cats[to_category_name])
            # get elements on source category
            dic = get_elements_by_category(doc, cats[from_category_name])
            # move elements
            return_value = move_elements_to_category(
                doc, dic, to_category_name, destination_cat_ids
            )
        else:
            return_value.update_sep(
                False,
                "Destination category: {} does not exist in file!".format(
                    to_category_name
                ),
            )
    else:
        return_value.update_sep(
            False,
            "Source category: {} does not exist in file!".format(from_category_name),
        )
    return return_value


def get_used_category_ids(doc):
    """
    Returns all category ids in a family which have an element assigned to them
    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of categories.
    :rtype: [Autodesk.Revit.DB.Category]
    """

    # get all elements in family
    dic = _sort_all_elements_by_category(doc)
    return dic.keys()



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit areas helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# required for ToList() call
import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

from duHast.Revit.Levels.levels import get_levels_in_model, get_level_elevation_by_name
from duHast.Revit.Common.parameter_get_utils import get_built_in_parameter_value
from duHast.Revit.Common.transaction import in_transaction
from duHast.Revit.Common.Geometry.curve import translate_curves_in_elevation
from duHast.Utilities.Objects import result as res

from duHast.Revit.Areas.areas import get_area_scheme_by_name

from Autodesk.Revit.DB import (
    BuiltInParameter,
    CurveElement,
    ElementId,
    ElementClassFilter,
    Transaction,
)


def get_area_lines_by_area_scheme_name(doc, scheme_name):
    """
    Returns a list of Revit model lines that are associated with a specific area scheme.

    Args:
        doc (Revit Document): The Revit document object.
        scheme_name (str): The name of the area scheme.

    Returns:
        list: A list of Revit elements representing the area lines associated with the specified area scheme.
    """
    return_value = []
    area_scheme = get_area_scheme_by_name(doc=doc, area_scheme_name=scheme_name)
    if area_scheme:
        filter = ElementClassFilter(CurveElement)
        dependent_element_ids = area_scheme.GetDependentElements(filter)
        for id in dependent_element_ids:
            element = doc.GetElement(id)
            return_value.append(element)
    return return_value


def sort_area_line_by_level_name(doc, area_lines):
    """
    Sorts the area lines based on their associated level names and returns a dictionary where the keys are the level names and the values are lists of area lines.

    Args:
        doc (Document): The current model document.
        area_lines (list): A list of area lines.

    Returns:
        dict: A dictionary where the keys are the level names and the values are lists of area lines.
    """
    levels = get_levels_in_model(doc)
    # build level id to level name dictionary
    level_names_by_id = {}
    for l in levels:
        level_names_by_id[l.Id] = l.Name
    # add edge case: line is not associated with a level ( maybe in a group?)
    level_names_by_id[ElementId.InvalidElementId] = 'This curve has no level associated'
    # build level name to area lines dictionary
    area_lines_by_Level_name = {}
    for area_line in area_lines:
        if level_names_by_id[area_line.LevelId] in area_lines_by_Level_name:
            area_lines_by_Level_name[level_names_by_id[area_line.LevelId]].append(
                area_line
            )
        else:
            area_lines_by_Level_name[level_names_by_id[area_line.LevelId]] = [area_line]
    return area_lines_by_Level_name


def delete_area_lines_by_level_name(
    doc, level_name, area_lines, transaction_manager=in_transaction
):
    """
    Deletes area lines in a Revit model based on their associated level name.

    Args:
        doc (Document): The current model document.
        level_name (str): The name of the level to delete area lines from.
        area_lines (list): A list of area lines to be deleted.
        transaction_manager (function, optional): A function that manages the transaction. Defaults to `in_transaction`.

    Returns:
        Result: A `Result` class instance containing the following attributes:
            - status (bool): True if the delete operation was successful, False otherwise.
            - message (str): A message indicating the status of the delete operation.
            - result (list): A list of remaining area lines after the delete operation.
    """
    return_value = res.Result()
    area_lines_return = area_lines
    # sort area lines by level
    lines_by_level_name = sort_area_line_by_level_name(doc=doc, area_lines=area_lines)
    if level_name in lines_by_level_name:
        element_ids_to_delete = []
        # get the ids to delete
        for element in lines_by_level_name[level_name]:
            element_ids_to_delete.append(element.Id)

        # define delete action
        def action():
            action_return_value = res.Result()
            try:
                element_ids_deleted = doc.Delete(
                    element_ids_to_delete.ToList[ElementId]()
                )
                action_return_value.update_sep(
                    True,
                    "Deleted {} area lines on level{}.".format(
                        len(element_ids_deleted), level_name
                    ),
                )
                action_return_value.result.append(element_ids_deleted)
            except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to delete {} area separation line(s) on level {} with exception: {}".format(
                        len(element_ids_to_delete), level_name, e
                    ),
                )
            return action_return_value

        tranny = Transaction(doc, "Deleting area lines on level: {}".format(level_name))
        result_delete = transaction_manager(tranny, action)
        return_value.update(result_delete)

        # return list of lines not deleted!
        if result_delete.status:
            lines_by_level_name.pop(level_name, None)
            area_lines_return = []
            for key, value in lines_by_level_name.items():
                area_lines_return.extend(value)
        return_value.result = [area_lines_return]

    else:
        return_value.update_sep(True, "No lines on level {} found.".format(level_name))
    return return_value


def create_new_area_outlines(doc, curves, view, transaction_manager=in_transaction):
    """
    Create new area separation lines in a specified view in Autodesk Revit.

    Args:
        doc (Revit Document): The Revit document in which the area separation lines will be created.
        curves (list): A list of curves representing the area separation lines.
        view (Revit View): The view in which the area separation lines will be created.
        transaction_manager (function, optional): A transaction manager function that wraps the creation of area separation lines in a Revit transaction. Default is `in_transaction`.

    Returns:
        Result: A `Result` object that contains the status of the transaction and any error or success messages.
    """
    return_value = res.Result()

    def action():
        action_return_value = res.Result()
        try:
            for t in curves:
                doc.Create.NewAreaBoundaryLine(view.SketchPlane, t, view)
            action_return_value.append_message(
                "Created new area separation line(s) in view {}".format(view.Name)
            )
        except Exception as e:
            action_return_value.update_sep(
                False,
                "Failed to create area separation line(s) in view {} with exception: {}".format(
                    view.Name, e
                ),
            )
        return action_return_value

    tranny = Transaction(
        doc, "Creating new area separation line(s) in view ".format(view.Name)
    )
    result_create = transaction_manager(tranny, action)
    return_value.update(result_create)
    return return_value


def copy_area_lines_to_level_name(
    doc, view, area_lines, transaction_manager=in_transaction
):
    """
    Copy area separation lines from one level to another in Autodesk Revit.

    Args:
        doc (Revit Document): The Revit document in which the area separation lines exist.
        view (Revit View): The view in which the area separation lines will be created.
        area_lines (list): A list of area separation lines to be copied.
        transaction_manager (function, optional): A transaction manager function that wraps the creation of area separation lines in a Revit transaction. Default is `in_transaction`.

    Returns:
        Result: A `Result` object that contains the status of the transaction and any error or success messages.
    """
    return_value = res.Result()
    try:
        # sort area lines by level
        lines_by_level_name = sort_area_line_by_level_name(
            doc=doc, area_lines=area_lines
        )
        level_name = get_built_in_parameter_value(
            view, BuiltInParameter.PLAN_VIEW_LEVEL
        )
        # make sure only one level was returned and its not the level to be copied to!
        if len(lines_by_level_name) != 1:
            raise ValueError(
                "Source area lines must all be on the same level. Found {} levels instead.".format(
                    len(lines_by_level_name)
                )
            )
        elif level_name in lines_by_level_name:
            raise ValueError(
                "Source {} and destination {} level are identical!".format(level_name)
            )
        else:
            # get curve geometry from model lines
            curves = [m_line.GeometryCurve for m_line in area_lines]
            # check if we got curves
            if any(curves):
                # get source elevation (the elevations the area lines are on)
                source_elevation = get_level_elevation_by_name(
                    doc=doc, level_name=lines_by_level_name.iterkeys().next()
                )
                # check if we got a source elevation
                if source_elevation != None:
                    # get target elevation
                    target_elevation = get_level_elevation_by_name(
                        doc=doc, level_name=level_name
                    )
                    # check if we got a target elevation
                    if target_elevation != None:
                        # translate curves from source level to target levels
                        translate_result = translate_curves_in_elevation(
                            doc=doc,
                            original_curves=curves,
                            source_elevation=source_elevation,
                            target_elevation=target_elevation,
                        )

                        # check if we got translated curves...
                        if translate_result.status:
                            # this is an Autodesk.Revit.DB.CurveArray
                            curve_array = translate_result.result[0]
                            create_result = create_new_area_outlines(
                                doc=doc,
                                curves=curve_array,
                                view=view,
                                transaction_manager=transaction_manager,
                            )
                            return_value.update(create_result)
                        else:
                            raise ValueError(translate_result.message)
                    else:
                        raise ValueError(
                            "Failed to get an elevation for level: {}".format(
                                level_name
                            )
                        )
                else:
                    raise ValueError(
                        "Failed to get an elevation for level: {}".format(
                            lines_by_level_name.iterkeys().next()
                        )
                    )
            else:
                raise ValueError("Failed to get curves from area lines.")

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to copy {} area separation line(s) to level(s) with exception: {}".format(
                len(area_lines),e
            ),
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of functions around exporting from Revit to nwc file format.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Revit.Views import views as rView
from duHast.Utilities.Objects import result as res
from duHast.Revit.Exports.export import build_export_file_name_from_view


def setup_nwc_custom_export_option(
    using_shared_coordinates,
    export_entire_model,
    export_links,
    split_model_by_level,
    export_parts,
    export_room_as_attributes,
    export_room_geometry,
    find_missing_materials,
    navis_parameters=rdb.NavisworksParameters.All,
    convert_element_properties=False,
):
    """
    Return an NWC Export Options object as per values past oin.
    :param using_shared_coordinates: True shared coordinates will be used, otherwise project internal
    :type using_shared_coordinates: bool
    :param export_entire_model: True entire model will be exported, otherwise specific view.
    :type export_entire_model: bool
    :param export_links: True: Revit links will also be exported, otherwise not.
    :type export_links: bool
    :param split_model_by_level: True: model elements will be split by level, otherwise not.
    :type split_model_by_level: bool
    :param export_parts: True parts will be exported, otherwise not.
    :type export_parts: bool
    :param export_room_as_attributes: True room properties will be exported (can be slow!), otherwise not.
    :type export_room_as_attributes: bool
    :param export_room_geometry: True room geometry will be exported, otherwise not.
    :type export_room_geometry: bool
    :param find_missing_materials: True exporter will attempt to find missing materials, otherwise not
    :type find_missing_materials: bool

    :param navis_parameters: Options which specifies the parameter conversion of Navisworks Exporter. Default value is All.
    :type navis_parameters: Autodesk.Revit.DB.NavisworksParameters

    :param convert_element_properties: True to convert element properties, false otherwise. Default value is false.
    :type convert_element_properties: bool

    :return: A Navisworks .nwc export option.
    :rtype: Autodesk.Revit.DB.NavisworksExportOptions
    """

    ex_nwc = rdb.NavisworksExportOptions()
    ex_nwc.Coordinates = (
        rdb.NavisworksCoordinates.Shared
        if using_shared_coordinates == True
        else rdb.NavisworksCoordinates.Internal
    )
    ex_nwc.ExportScope = (
        rdb.NavisworksExportScope.Model
        if export_entire_model == True
        else rdb.NavisworksExportScope.View
    )
    ex_nwc.ExportLinks = export_links
    ex_nwc.DivideFileIntoLevels = split_model_by_level
    ex_nwc.ExportParts = export_parts
    ex_nwc.ExportRoomAsAttribute = export_room_as_attributes
    ex_nwc.ExportRoomGeometry = export_room_geometry
    ex_nwc.FindMissingMaterials = find_missing_materials
    ex_nwc.ConvertElementProperties = convert_element_properties
    ex_nwc.Parameters = navis_parameters

    return ex_nwc


def setup_nwc_default_export_option_shared_by_view():
    """
    Return an NWC Export Options object with shared coordinates, export by View.
    :return: A Navisworks .nwc export option.
    :rtype: Autodesk.Revit.DB.NavisworksExportOptions
    """

    return setup_nwc_custom_export_option(
        using_shared_coordinates=True,
        export_entire_model=False,
        export_links=False,
        split_model_by_level=True,
        export_parts=True,
        export_room_as_attributes=False,
        export_room_geometry=False,
        find_missing_materials=False,
    )


def export_to_nwc(doc, nwc_export_option, directory_path, file_name):
    """
    Exports either the entire model or a view to NWC (Navisworks Cache) format.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param nwc_export_option: A Navisworks .nwc export option.
    :type nwc_export_option: Autodesk.Revit.DB.NavisworksExportOptions
    :param directory_path: The directory path to where the export is being saved.
    :type directory_path: str
    :param file_name: The file name under which the export is being saved.
    :type file_name: str

    :return: Result class instance.
        - `result.status` (bool): False if an exception occurred, otherwise True (export status).
        - `result.message` (str): Contains the fully qualified file path of the exported file.

    On exception:
        - `result.status` (bool): False.
        - `result.message` (str): Contains the exception message.

    :rtype: :class:`.Result`
    """

    # nwc export does not need to run in a transaction
    return_value = res.Result()
    try:
        # export to NWC
        doc.Export(directory_path, file_name, nwc_export_option)
        return_value.update_sep(
            True, "Exported: {}".format(directory_path + "\\" + str(file_name))
        )
        # needs to be a list in a list to stay together when combined with previous results in the update status result code
        return_value.result = [[directory_path, file_name]]
    except Exception as e:
        return_value.update_sep(
            False,
            "Script Exception: Failed to export to NWC with exception: {}".format(e),
        )
    return return_value


def export_model_to_nwc(doc, nwc_export_option, directory_path, file_name):
    """
    Exports the entire Revit model to NWC (Navisworks Cache) format.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param nwc_export_option: A Navisworks .nwc export option.
    :type nwc_export_option: Autodesk.Revit.DB.NavisworksExportOptions
    :param directory_path: The directory path where the export is being saved.
    :type directory_path: str
    :param file_name: The file name under which the export is being saved.
    :type file_name: str

    :return: Result class instance.
        - `result.status` (bool): False if an exception occurred, otherwise True (export status).
        - `result.message` (str): Contains the fully qualified file path of the exported file.

    On exception:
        - `result.status` (bool): False.
        - `result.message` (str): Contains the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    return_value_by_model = export_to_nwc(
        doc, nwc_export_option, directory_path, file_name
    )
    return_value.update(return_value_by_model)
    return return_value


def export_3d_views_to_nwc(
    doc,
    view_filter,
    nwc_export_option,
    directory_path,
    do_something_with_view_name=None,
):
    """
    Exports 3D views matching a filter (view starts with) to NWC (Navisworks Cache) format.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view_filter: String the view name is to start with if it is to be exported. (Both view name and string are set to lowercase for comparison)
    :type view_filter: str
    :param nwc_export_option: A Navisworks .nwc export option.
    :type nwc_export_option: Autodesk.Revit.DB.NavisworksExportOptions
    :param directory_path: The directory path where the export is being saved.
    :type directory_path: str
    :param do_something_with_view_name: A function that takes the view name as an argument and processes it. The modified view name is then used as the actual file name, defaults to None, which uses the view name unchanged as the export file name.
    :type do_something_with_view_name: function, optional

    :return: Result class instance.
        - `result.status` (bool): False if an exception occurred, otherwise True (export status).
        - `result.message` (str): Contains the fully qualified file path of the exported file.

    On exception:
        - `result.status` (bool): False.
        - `result.message` (str): Contains the exception message.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    views_to_export = []
    # get all 3D views in model and filter out views to be exported
    views = rView.get_views_of_type(doc, rdb.ViewType.ThreeD)
    for v in views:
        if v.Name.lower().startswith(view_filter.lower()):
            views_to_export.append(v)
    # export those views one by one
    if len(views_to_export) > 0:
        for export_view in views_to_export:
            return_value_by_view = res.Result()
            # store view ID in export option
            nwc_export_option.ViewId = export_view.Id
            file_name = (
                build_export_file_name_from_view(export_view.Name, view_filter, ".nwc")
                if do_something_with_view_name == None
                else do_something_with_view_name(export_view.Name)
            )
            return_value_by_view = export_to_nwc(
                doc, nwc_export_option, directory_path, file_name
            )
            return_value.update(return_value_by_view)
    else:
        return_value.update_sep(
            True, "NWC Export: No 3D views found matching filter...nothing was exported"
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit geometry extraction helper functions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr

from duHast.Revit.Common.Geometry.points import (
    are_points_identical,
    check_duplicate_point,
    flatten_xyz_point,
    get_point_as_doubles,
    get_point_as_string,
)
from duHast.Utilities.compare import is_close

clr.AddReference("System.Core")
from System import Linq
from System import Math

clr.ImportExtensions(Linq)

from Autodesk.Revit.DB import BoundingBoxXYZ, PlanarFace, XYZ


def get_edge_as_string(edge):
    """
    Returns a revit edge as a string.

    :param edge: A revit edge.
    :type edge: Autodesk.Revit.DB.Edge

    :return: String where each row represents a point on the edge.
    :rtype: str
    """

    returnValue = ""
    for p in edge.Tessellate():
        returnValue = returnValue + "\n" + get_point_as_string(p)
    return returnValue


def UV_pt_list_from_crv_list(curve_list):
    """
    Returns a list of UV points from a list of curves
    :param curve_list: A list of curves
    :param type: list
    :return: A list of UV points
    :rtype: list
    """
    pt_list = []

    for crv in curve_list:
        st_pt = crv.GetEndPoint(0)
        pt_list.append((st_pt.X, st_pt.Y))

    return pt_list


def point_in_polygon(point, polygon):
    """
    Returns True if a point is inside a polygon, and False if it is not
    :param point: The point to check
    :type point: list, tuple, or set
    :param polygon: The polygon to check
    :type polygon: list, tuple, or set
    :return: True if the point is inside the polygon, False if it is not
    :rtype: bool
    """

    num_of_polygon_sides = len(polygon)
    is_inside = False

    pt_1_x, pt_1_y = polygon[0]
    for i in range(num_of_polygon_sides + 1):
        pt_2_x, pt_2_y = polygon[i % num_of_polygon_sides]
        if point[1] > min(pt_1_y, pt_2_y):
            if point[1] <= max(pt_1_y, pt_2_y):
                if point[0] <= max(pt_1_x, pt_2_x):
                    if pt_1_y != pt_2_y:
                        x_intersection = (point[1] - pt_1_y) * (pt_2_x - pt_1_x) / (
                            pt_2_y - pt_1_y
                        ) + pt_1_x
                    if pt_1_x == pt_2_x or point[0] <= x_intersection:
                        is_inside = not is_inside
        pt_1_x, pt_1_y = pt_2_x, pt_2_y

    return is_inside


# ---------------------------- math utility ----------------------------


def flatten_xyz_point_list(polygon):
    """
    Flattens a list of XYZ points to a list of UV points by omitting the Z value of each XYZ.

    https://thebuildingcoder.typepad.com/blog/2008/12/2d-polygon-areas-and-outer-loop.html

    :param polygon: list of XYZ points
    :type polygon: list Autodesk.Revit.DB.XYZ

    :return: list of UV points
    :rtype: list Autodesk.Revit.DB.UV
    """

    z = polygon[0].Z
    a = []
    for p in polygon:
        if is_close(p.Z, z):
            # print("expected horizontal polygon" )
            pass
        a.append(flatten_xyz_point(p))
    return a


def flatten_xyz_point_list_of_lists(polygons):
    """
    Flattens a list lists of XYZ points to a list of lists of UV points by omitting the Z value of each XYZ.

    https://thebuildingcoder.typepad.com/blog/2008/12/2d-polygon-areas-and-outer-loop.html

    :param polygon: list of lists of XYZ points
    :type polygon: list [list Autodesk.Revit.DB.XYZ]

    :return: list of lists of UV points
    :rtype: list [list Autodesk.Revit.DB.UV]
    """

    z = polygons[0][0].Z
    a = []
    for polygon in polygons:
        if is_close(polygon[0].Z, z):
            # print("expected horizontal polygon" )
            pass
        a.append(flatten_xyz_point_list(polygon))
    return a


def get_coordinate_system_translation_and_rotation(doc):
    """
    Returns the rotation as a 3 x 3 matrix and the translation as a 1 x 3 matrix of the shared coordinate system active in document.

    :param doc: _description_
    :type doc: _type_

    :return: 3 x 3 matrix describing rotation, 1 x 3 matrix describing translation
    :rtype: list (3) [list(3) int], list [int]
    """

    projectLocationActive = doc.ActiveProjectLocation
    # get the inverse because we need to go back to origin
    totalTransform = projectLocationActive.GetTotalTransform().Inverse
    nBasisX = get_point_as_doubles(totalTransform.BasisX)
    nBasisY = get_point_as_doubles(totalTransform.BasisY)
    nBasisZ = get_point_as_doubles(totalTransform.BasisZ)
    nOrigin = get_point_as_doubles(totalTransform.Origin)
    return [nBasisX, nBasisY, nBasisZ], nOrigin


# --------------------------------------- is point in polygon ---------------------------------------
# from  https://thebuildingcoder.typepad.com/blog/2010/12/point-in-polygon-containment-algorithm.html
# ---------------------------------------------------------------------------------------------------


def get_quadrant(vertex, p):
    """
    Determines the quadrant of a polygon vertex relative to the test point.

    https://thebuildingcoder.typepad.com/blog/2010/12/point-in-polygon-containment-algorithm.html

    :param vertex: Revit UV element describing a vertex
    :type vertex: Autodesk.Revit.DB.UV
    :param p: Revit UV point
    :type p: Autodesk.Revit.DB.UV

    :return: An integer of range 0 - 4 describing the quadrant.
    :rtype: int
    """

    returnValue = None
    if vertex.U > p.U:
        if vertex.V > p.V:
            returnValue = 0
        else:
            returnValue = 3
    else:
        if vertex.V > p.V:
            returnValue = 1
        else:
            returnValue = 2
    return returnValue


def x_intercept(p, q, y):
    """
    Determine the X intercept of a polygon edge with a horizontal line at the Y value of the test point.

    https://thebuildingcoder.typepad.com/blog/2010/12/point-in-polygon-containment-algorithm.html

    :param p: Revit UV point
    :type p: Autodesk.Revit.DB.UV
    :param q: Revit UV point
    :type q: Autodesk.Revit.DB.UV
    :param y: _description_
    :type y: double

    :return: _description_
    :rtype: double
    """
    if 0 != (p.V - q.V):
        # print('unexpected horizontal segment')
        pass

    return q.U - ((q.V - y) * ((p.U - q.U) / (p.V - q.V)))


def adjust_delta(delta, vertex, next_vertex, p):
    """
    https://thebuildingcoder.typepad.com/blog/2010/12/point-in-polygon-containment-algorithm.html

    :param delta: _description_
    :type delta: _type_
    :param vertex: _description_
    :type vertex: _type_
    :param next_vertex: _description_
    :type next_vertex: _type_
    :param p: _description_
    :type p: _type_
    :return: _description_
    :rtype: _type_
    """

    returnValue = delta
    # make quadrant deltas wrap around:
    if delta == 3:
        returnValue = -1
    elif delta == -3:
        returnValue = 1
    # check if went around point cw or ccw:
    elif delta == 2:
        returnValue = 2
    elif delta == -2:
        if x_intercept(vertex, next_vertex, p.V) > p.U:
            returnValue = -delta
    return returnValue


def is_point_within_polygon(polygon, point):
    """
    Checks whether a point is within a polygon.

    https://thebuildingcoder.typepad.com/blog/2010/12/point-in-polygon-containment-algorithm.html
    odd number of windings rule:
    if (angle & 4) return INSIDE else return OUTSIDE
    non-zero winding rule:
    if (angle != 0) return INSIDE else return OUTSIDE

    :param polygon: A polygon
    :type polygon: list of Autodesk.Revit.DB.UV
    :param point: A point
    :type point: Autodesk.Revit.DB.UV

    :return: Refer winding rules above.
    :rtype: bool
    """

    # initialize
    quad = get_quadrant(polygon[0], point)
    angle = 0

    # loop on all vertices of polygon
    next_quad = 0
    delta = 0
    n = len(polygon)
    for i in range(n):
        vertex = polygon[i]
        next_vertex = None
        if i + 1 < n:
            next_vertex = polygon[i + 1]
        else:
            next_vertex = polygon[0]
        # calculate quadrant and delta from last quadrant
        next_quad = get_quadrant(next_vertex, point)
        delta = next_quad - quad
        delta = adjust_delta(delta, vertex, next_vertex, point)
        # add delta to total angle sum
        angle = angle + delta
        # increment for next step
        quad = next_quad
    """ 
    odd number of windings rule:
    if (angle & 4) return INSIDE; else return OUTSIDE;
    non-zero winding rule:
    if (angle != 0) return INSIDE; else return OUTSIDE;
    """
    # complete 360 degrees (angle of + 4 or -4 ) means inside
    return (angle == +4) or (angle == -4)


# --------------------------------------- END --------------------------------------------------


def get_signed_polygon_area(uv_points):
    """
    Calculates the area of a signed UV polygon.

    https://thebuildingcoder.typepad.com/blog/2008/12/2d-polygon-areas-and-outer-loop.html

    :param uv_points: list of points defining the polygon.
    :type uv_points: list Autodesk.Revit.DB.UV

    :return: The area of the polygon.
    :rtype: double
    """

    n = len(uv_points)
    sum = uv_points[0].U * (uv_points[1].V - uv_points[n - 1].V)
    for i in range(1, n - 1):
        sum += uv_points[i].U * (uv_points[i + 1].V - uv_points[i - 1].V)
    sum += uv_points[n - 1].U * (uv_points[0].V - uv_points[n - 2].V)
    return 0.5 * sum


def convert_edge_arrays_into_list_of_points(edge_arrays):
    """
    Converts an edge array into a list of list of revit XYZ points.

    :param edge_arrays: A revit edge array.
    :type edge_arrays: Autodesk.Revit.DB.EdgeArrayArray ( no not a spelling mistake :) )

    :return: A List of list of revit XYZ points.
    :rtype: list of list Autodesk.Revit.DB.XYZ
    """

    polygons = []
    for loop in edge_arrays:
        vertices = []
        q = None
        (first_point) = True
        for edge in loop:
            points = edge.Tessellate()
            if first_point:
                q = points[0]
            n = len(points)
            for i in range(n - 1):
                vertices.append(points[i])
            (first_point) = False
        # close the loop by ending with first point...not required(?)
        # vertices.append(q)
        polygons.append(vertices)
    return polygons


def get_edge_points(edge):
    """
    Retrieves the revit XYZ points defining an edge (curves get tessellated!)

    :param edge: An edge of a solid.
    :type edge: Autodesk.Revit.DB.Edge

    :return: A list of revit XYZ points.
    :rtype: list Autodesk.Revit.DB.XYZ
    """

    points = []
    for p in edge.Tessellate():
        points.append(p)
    return points


def check_duplicate_edge(edges, edge):
    """
    Checks whether a collection contains a given edge and returns True if that is the case.

    :param edges: List of edges toi check against.
    :type edges: list of Autodesk.Revit.DB.Edge
    :param edge: An edge
    :type edge: Autodesk.Revit.DB.Edge

    :return: True if edge is already in collection, otherwise False.
    :rtype: bool
    """

    flag_over_all = False
    comp_points = get_edge_points(edge)
    if len(edges) > 0:
        for e in edges:
            list_points = get_edge_points(e)
            if len(comp_points) == len(list_points):
                edge_same_flag = True
                for p1 in comp_points:
                    if check_duplicate_point(list_points, p1) == False:
                        edge_same_flag = False
                if edge_same_flag:
                    flag_over_all = True
                    break
            else:
                pass
    else:
        pass  # flag is already False
    return flag_over_all


def check_solid_is_zero_height(solid):
    """
    Checks if points making up a solid have multiple Z values.

    :param solid: A revit solid object.
    :type solid: Autodesk.Revit.DB.Solid

    :return: False if points collection of a solid has multiple z values, True if only one Z value in points collection (2D solid...??)
    :rtype: bool
    """

    return check_edges_are_zero_height(solid.Edges)


def check_edges_are_zero_height(edges):
    """
    Checks if points making up a edges have multiple Z values.

    :param edges: A list of edges.
    :type edges: list Autodesk.Revit.DB.Edge

    :return: False if points collection of a edges has multiple z values, True if only one Z value in points collection (2D edges...??)
    :rtype: bool
    """

    lowest_z = 0.0
    counter = 0
    for edge in edges:
        for p in edge.Tessellate():
            if counter == 0:
                lowest_z = p.Z
            else:
                if is_close(p.Z, lowest_z) == False:
                    # found multiple z
                    return False
            counter = counter + 1
    return True


def get_lowest_z_from_solid(solid):
    """
    Gets the lowest Z value in a solids point collection.

    :param solid: A solid.
    :type solid: Autodesk.Revit.DB.Solid

    :return: The lowest Z value of any point in the solids vertex collection.
    :rtype: double
    """

    return get_lowest_z_from_edges_point_collection(solid.Edges)


def get_lowest_z_from_edges_point_collection(edges):
    """
    Gets the lowest Z value in a edges collection

    :param edges: A list of edges.
    :type edges: list Autodesk.Revit.DB.Edge

    :return: The lowest Z value of any point in the edges vertex collection.
    :rtype: double
    """

    lowest_z = 0.0
    counter = 0
    for edge in edges:
        for p in edge.Tessellate():
            # make sure lowest z is initialized with the first point z values
            if counter == 0:
                lowest_z = p.Z
            else:
                if is_close(p.Z, lowest_z) == False and p.Z < lowest_z:
                    lowest_z = p.Z
            counter = counter + 1
    return lowest_z


def edges_are_connected(edge1, edge2):
    """
    Checks whether edges are connected by comparing their points. If there is an identical point in both\
        then they are connected.

    Revit solids dont have crossing edges!

    :param edge1: An edge.
    :type edge1: Autodesk.Revit.DB.Edge
    :param edge2: Another edge.
    :type edge2: Autodesk.Revit.DB.Edge

    :return: _description_
    :rtype: bool
    """

    for p1 in edge1.Tessellate():
        for p2 in edge2.Tessellate():
            if are_points_identical(p1, p2):
                return True
    return False


def get_faces_sorted_by_area_from_solid(solid):
    """
    Returns all faces from a solid sorted descending from biggest to smallest by area.

    :param solid: A solid.
    :type solid: Autodesk.Revit.DB.Solid

    :return: A list of faces, sorted biggest to smallest by area.
    :rtype: list Autodesk.Revit.DB.Face
    """

    fl = []
    for face in solid.Faces:
        fl.append(face)
    return sorted(fl, key=lambda x: x.Area, reverse=True)


def pair_faces_by_area(faces):
    """
    Returns a list of lists of face pairs, where a nested list contains faces with the same measured area.

    Sample would be a ceiling solid. The top and bottom face of that ceiling would be an area pair.

    :param faces: A list of faces.
    :type faces: list Autodesk.Revit.DB.Face

    :return: A list of lists of faces.
    :rtype: list of list Autodesk.Revit.DB.Face
    """

    return_value = []
    # duplicate faces list since it will be manipulated
    copy_faces = list(faces)
    flag = True
    while flag:
        # loop over faces and try to find some with the same area measured
        same_area = [copy_faces[0]]
        f = copy_faces[0]
        for i in range(1, len(copy_faces)):
            if is_close(f.Area, copy_faces[i].Area):
                same_area.append(copy_faces[i])
        return_value.append(same_area)
        # removes faces accounted for
        for a in same_area:
            copy_faces.remove(a)
        if len(copy_faces) < 2:
            flag = False
    return return_value


def get_faces_with_lowest_z_from_pairs(face_pairs):
    """
    Gets the face with the lowest Z value from list of faces.

    :param face_pairs: A list of lists of face pairs, where a nested list contains faces with the same measured area.
    :type face_pairs: list of list Autodesk.Revit.DB.Face

    :return: A list of faces.
    :rtype: list Autodesk.Revit.DB.Face
    """

    faces = []
    for faceP in face_pairs:
        lowestZ = 0.0
        counter = 0
        current_face = None
        for face in faceP:
            current_z = get_lowest_z_from_edges_point_collection(face.EdgeLoops[0])
            if counter == 0:
                current_face = face
                lowestZ = current_z
                counter = counter + 1
            else:
                if current_z < lowestZ:
                    current_face = face
                    lowestZ = current_z
        faces.append(current_face)
    return faces


def get_unique_horizontal_faces(faces):
    """
    Filters out any horizontal faces from list of faces past in.

    Will also further filter by: faces with the same area only the face with the lower Z value will be returned.
    Note: works only on planar faces

    TODO: It could be way simpler just to check for the face with a negative face normal Z value...

    :param faces: A list of faces.
    :type faces: list Autodesk.Revit.DB.Face

    :return: A list of faces.
    :rtype: list Autodesk.Revit.DB.Face
    """

    faces_horizontal = []
    for f in faces:
        # non planar faces are ignored for the moment...
        if type(f) is PlanarFace:
            if f.FaceNormal.Z != 0.0:
                faces_horizontal.append(f)

    faces_filtered = []
    if len(faces_horizontal) > 1:
        # pair faces by area
        paired_faces = pair_faces_by_area(faces_horizontal)
        # get faces with lowest Z value for each pair
        faces_filtered = get_faces_with_lowest_z_from_pairs(paired_faces)
    return faces_filtered


def is_loop_within_other_loop_but_not_reference_loops(
    exterior_loop, other_loop, hole_loops
):
    """
    Checks whether any of the other loops is within the exterior loop and if so\
        if it is not also within one of the holeLoops ...that would be an island
    
    :param exterior_loop: A polygon loop describing the external boundary of a face.
    :type exterior_loop: list of Autodesk.Revit.DB.UV
    :param other_loop: A polygon loop which is to be checked as to whether it is within the exterior loop and the any hole loops.
    :type other_loop: list of Autodesk.Revit.DB.UV
    :param hole_loops: A list of named tuples containing .loop property which is a list of UV points forming a polygon which have been identified\
         as creating a hole in the exteriorLoop.
    :type hole_loops: namedtuple('uvLoop', 'loop area id threeDPoly')\
        .Loop is a list of UV points defining a polygon loop
        .area is a double describing the polygon area
        .id is an integer
        .threeDPoly is an edge loop
    
    :return: True if within exterior loop but not within hole loops, otherwise False.
    :rtype: bool
    """

    return_value = False
    # get any point on the loop to check...if it is within the other loop then the entire loop is within the other loop
    # since revit does not allow for overlapping sketches
    point = other_loop[0]
    # check whether point is within the other polygon loop
    if is_point_within_polygon(exterior_loop, point):
        return_value = True
        # check whether this point is within the polygon loops identified as holes
        # if so it is actually an island and will be accounted for separately
        if len(hole_loops) > 0:
            for hole_loop in hole_loops:
                if is_point_within_polygon(hole_loop.loop, point):
                    return_value = False
                    break
    return return_value


def build_loops_dictionary(loops):
    """
    Will return a dic where:

    - key is the outer loop of a polygon id
    - values is a list of tuples describing holes in the key polygon

    :param loops: A list of named tuples describing polygons.\
        .Loop is a list of UV points defining a polygon loop
        .area is a double describing the polygon area
        .id is an integer
        .threeDPoly is an edge loop
    :type loops: list[namedtuple('uvLoop', 'loop area id threeDPoly')]

    :return: A dictionary.
    :rtype: dic {int: list[namedtuple('uvLoop', 'loop area id threeDPoly')]}
    """

    # duplicate list since I'am about to manipulate it...
    copy_loops = list(loops)
    flag = False
    return_value = {}
    counter = 0
    # check if there is more then one loop  to start with
    if len(copy_loops) > 1:
        loop_flag = True
        while loop_flag:
            # add the biggest loop as exterior to dictionary (first one in list)
            key = copy_loops[0].id
            return_value[key] = []
            # assign loop to be checked for holes
            reference_loop = copy_loops[0]
            # remove the first exterior loop from list of loops
            copy_loops.pop(0)
            # loop over remaining loops and work out which ones are holes ... if any
            for loop in copy_loops:
                # build list of hole loops already known belonging to this exterior loop
                hole_loops = []
                for geo_loop in return_value[key]:
                    hole_loops.append(geo_loop)
                # check whether this is another hole loop
                flag = is_loop_within_other_loop_but_not_reference_loops(
                    reference_loop.loop, loop.loop, hole_loops
                )
                if flag:
                    return_value[key].append(loop)
            # remove loops identified as holes from overall list as to avoid double counting
            for hole in return_value[key]:
                counter = counter + 1
                copy_loops.remove(hole)
            # check whether all loops are accounted for
            if len(copy_loops) == 0:
                loop_flag = False
    elif len(copy_loops) == 1:
        key = copy_loops[0].id
        # only one exterior loop exists, no interior loops
        return_value[key] = []
    return return_value


def negate_vector(vector):
    """
    Negates the direction of a given vector
    :param vector: The vector to negate
    :type vector: list or tuple
    :return: The negated vector
    :rtype: list
    """
    return [-x for x in vector]


def merge_bounding_box_xyz(bounding_box_xyz_0, bounding_box_xyz_1):
    """
    Merges two bounding boxes into one.
    :param bounding_box_xyz_0: The first bounding box
    :type bounding_box_xyz_0: BoundingBoxXYZ
    :param bounding_box_xyz_1: The second bounding box
    :type bounding_box_xyz_1: BoundingBoxXYZ

    :return: The merged bounding box
    :rtype: BoundingBoxXYZ
    """

    merged_result = BoundingBoxXYZ()
    merged_result.Min = XYZ(
        Math.Min(bounding_box_xyz_0.Min.X, bounding_box_xyz_1.Min.X),
        Math.Min(bounding_box_xyz_0.Min.Y, bounding_box_xyz_1.Min.Y),
        Math.Min(bounding_box_xyz_0.Min.Z, bounding_box_xyz_1.Min.Z),
    )

    merged_result.Max = XYZ(
        Math.Max(bounding_box_xyz_0.Max.X, bounding_box_xyz_1.Max.X),
        Math.Max(bounding_box_xyz_0.Max.Y, bounding_box_xyz_1.Max.Y),
        Math.Max(bounding_box_xyz_0.Max.Z, bounding_box_xyz_1.Max.Z),
    )
    return merged_result



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to spot annotation. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from Autodesk.Revit.DB import (
    BuiltInCategory,
    BuiltInParameter,
    ElementId,
    FilteredElementCollector,
    SpotDimensionType,
)
from duHast.Revit.Annotation.arrow_heads import (
    get_arrow_head_ids_from_type,
    ARROWHEAD_PARAS_SPOT_DIMS,
)
from duHast.Revit.Common.common import get_ids_from_element_collector
from duHast.Revit.Common.parameter_get_utils import get_built_in_parameter_value


def get_all_spot_dim_types(doc):
    """
    Gets all spot Dim types in the model

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector of spot dimension types
    :rtype: Autodesk.Revit.DB.FilteredElementCollector of spot dimension types
    """

    return FilteredElementCollector(doc).OfClass(SpotDimensionType)


def get_symbol_ids_from_spot_types(doc):
    """
    Gets all family symbol ids used as a symbol from all spot elevation type definitions and spot coordinate type definitions.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing family symbols
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    dim_ts = get_all_spot_dim_types(doc)
    for t in dim_ts:
        id = get_built_in_parameter_value(t, BuiltInParameter.SPOT_ELEV_SYMBOL)
        if id not in ids and id != ElementId.InvalidElementId and id != None:
            ids.append(id)
    return ids


def get_all_spot_elevation_symbols_in_model(doc):
    """
    Gets all family symbols of category Spot Elevation Symbol in model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    col = FilteredElementCollector(doc).OfCategory(BuiltInCategory.OST_SpotElevSymbols)
    return col


def get_all_spot_elevation_symbol_ids_in_model(doc):
    """
    Gets all family symbol ids of category Spot Elevation Symbol in model
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing family symbols
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    col = get_all_spot_elevation_symbols_in_model(doc)
    ids = get_ids_from_element_collector(col)
    return ids


def get_spot_type_arrow_head_ids(doc):
    """
    returns all arrow head ids used in text types in a model
    """
    used_ids = get_arrow_head_ids_from_type(
        doc, get_all_spot_dim_types, ARROWHEAD_PARAS_SPOT_DIMS
    )
    return used_ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions relating to comparing:. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

- two values
- whether a text value starts or does not start with a given text value

"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


def does_not_equal(valueOne, valueTwo):
    """
    Returns True if valueOne does not match valueTwo.
    :param valueOne: a value
    :type valueOne: var
    :param valueTwo: another value
    :type valueTwo: var
    :return: True if valueOne does not match valueTwo, otherwise False
    :rtype: bool
    """

    if valueOne != valueTwo:
        return True
    else:
        return False


def does_equal(valueOne, valueTwo):
    """
    Returns True if valueOne does match valueTwo.
    :param valueOne: a value
    :type valueOne: var
    :param valueTwo: another value
    :type valueTwo: var
    :return: True if valueOne does match valueTwo, otherwise False
    :rtype: bool
    """

    if valueOne == valueTwo:
        return True
    else:
        return False


def one_start_with_two(valueOne, valueTwo):
    """
    Returns True if valueOne starts with valueTwo.
    :param valueOne: a value
    :type valueOne: str
    :param valueTwo: another value
    :type valueTwo: str
    :return: True if valueOne starts with valueTwo, otherwise False
    :rtype: bool
    """

    if valueOne.startswith(valueTwo):
        return True
    else:
        return False


def two_does_not_start_with_one(valueOne, valueTwo):
    """
    Returns True if valueTwo does not starts with valueOne.
    :param valueOne: a value
    :type valueOne: str
    :param valueTwo: another value
    :type valueTwo: str
    :return: True if valueTwo does not starts with valueOne, otherwise False
    :rtype: bool
    """

    if valueTwo.startswith(valueOne):
        return False
    else:
        return True


def is_close(a, b, rel_tol=1e-09, abs_tol=0.0):
    """
    Compares two floats with a tolerance. Returns True if they are close enough, otherwise False

    refer to: https://stackoverflow.com/questions/5595425/what-is-the-best-way-to-compare-floats-for-almost-equality-in-python

    :param a: A float
    :type a: float
    :param b: A float
    :type b: float
    :param rel_tol: Relative tolerance used to compare the two floats, defaults to 1e-09
    :type rel_tol: float, optional
    :param abs_tol: Absolute tolerance used to compare the two floats, defaults to 0.0
    :type abs_tol: float, optional

    :return: Returns True if they are close enough to be considered equal, otherwise False
    :rtype: bool
    """

    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)



from duHast.UI.Objects.ViewModelBase import ViewModelBase
from duHast.UI.Objects.Command import Command
from duHast.UI.Objects.file_item import MyFileItem
from duHast.UI.file_list import get_revit_files_for_processing


from System.Collections.ObjectModel import ObservableCollection
import System


class ViewModel(ViewModelBase):

    PropertyChanged = None

    def __init__(
        self,
        window,
        directory_source_path,
        directory_destination_path,
        source_file_extension,
        number_of_task_files,
        include_sub_dirs,
        filter_rules,
        filter_is_and,
    ):

        ViewModelBase.__init__(self)
        
        # check past in variable types
        if(not(isinstance(filter_rules,list))):
            raise TypeError ("filter_rules must be a list")
        
        self.window = window
        self.debug = []

        # set up an observable collection for files found 
        self.revit_files = ObservableCollection[MyFileItem]()
        # set up an observable collection for files filtered
        self.filtered_revit_files = ObservableCollection[MyFileItem]()
        
        # variable to contain the selected files in the data grid
        self.selected_files = ObservableCollection[MyFileItem]()

        self.source_path = directory_source_path
        self.destination_path = directory_destination_path
        self.source_file_extension = source_file_extension
        self.number_of_task_files = number_of_task_files
        self.include_sub_dirs = include_sub_dirs
        self.filter_rules = filter_rules
        
        # build filter text string from filter rules
        if(len(filter_rules)>1):
            self.filter_text = ";".join(self.filter_rules)
        elif(len(filter_rules)==1):
            self.filter_text = filter_rules[0]
        else:
            self.filter_text=""
        
        # set radio buttons defining filter type ( AND or OR)
        self.filter_is_and = filter_is_and
        self.filter_is_or = not (filter_is_and)

        # hook up ok and cancel buttons
        self.BtnOkCommand = Command(self.button_ok)
        self.BtnCancelCommand = Command(self.button_cancel)

        # get files
        self._get_files()
        
        # filter data of data grid view
        self.TextBox_Filter_TextChanged()

        # set the dialog result property to true as a default
        self.DialogResult = True

    @property
    def TextBoxText_SourcePath(self):
        return self.source_path

    @TextBoxText_SourcePath.setter
    def TextBoxText_SourcePath(self, value):
        if self.source_path != value:
            self.source_path = value
            self.OnPropertyChanged("TextBoxText_SourcePath")
            # this appears to be a bit of hack since I cant seem to be able to hook up an event handler to the text changed event via XAML
            self.text_box_source_path_change()

    @property
    def TextBoxText_Filter(self):
        return self.filter_text

    @TextBoxText_Filter.setter
    def TextBoxText_Filter(self, value):
        if self.filter_text != value:
            self.filter_text = value
            self.OnPropertyChanged("TextBoxText_Filter")
            # this appears to be a bit of hack since I cant seem to be able to hook up an event handler to the text changed event via XAML
            self.TextBox_Filter_TextChanged()

    @property
    def RadioButton_FilterAnd(self):
        return self.filter_is_and
    
    @RadioButton_FilterAnd.setter
    def RadioButton_FilterAnd(self, value):
        if self.filter_is_and != value:
            self.filter_is_and = value
            self.OnPropertyChanged("RadioButton_FilterAnd")
            # this appears to be a bit of hack since I cant seem to be able to hook up an event handler to the radio button changed event via XAML
            self.TextBox_Filter_TextChanged()
            
    @property
    def RadioButton_FilterOr(self):
        return self.filter_is_or
    
    @RadioButton_FilterOr.setter
    def RadioButton_FilterOr(self, value):
        if self.filter_is_or != value:
            self.filter_is_or = value
            self.OnPropertyChanged("RadioButton_FilterOr")
            # this appears to be a bit of hack since I cant seem to be able to hook up an event handler to the radio button changed event via XAML
            self.TextBox_Filter_TextChanged()

    def convert_list_to_observable_collection(self, list_of_items, target_list):
        """
        Used transfer items from a standard python list to another list ( of a .net type)

        Args:
            list_of_items (_type_): _description_
            target_list (_type_): _description_
        """
        target_list.Clear()  # Clear existing items if any
        for item in list_of_items:
            target_list.Add(item)

    def OnPropertyChanged(self, name):
        if self.PropertyChanged:
            self.PropertyChanged(
                self, System.ComponentModel.PropertyChangedEventArgs(name)
            )

    def _apply_and_to_file_data(self):
        """
        Filters file data assuming logical AND rule

        Returns:
            _type_: _description_
        """
        filtered_revit_files_intermediate = []
        # filter file list
        for file in self.revit_files:
            all_match_by_file = True
            # check each filter
            for filter in self.filter_rules:
                if filter not in file.name:
                    # all filters need to match...this is not the case
                    # set flag accordingly
                    all_match_by_file = False
                    break
            # only add file if all filters returned true
            if( all_match_by_file ):
                filtered_revit_files_intermediate.append(file)
            
        return filtered_revit_files_intermediate
    
    def _apply_or_to_file_data(self):
        """
        Filters file data assuming logical OR rule

        Returns:
            _type_: _description_
        """
        filtered_revit_files_intermediate = []
        for file in self.revit_files:
            any_match_by_file = False
            # check each filter
            for filter in self.filter_rules:
                if filter in file.name:
                    # just one filters need to match...this is a match
                    # set flag accordingly
                    any_match_by_file = True
                    # one match is enough...move on
                    break
            # add file if any filters returned true
            if( any_match_by_file ):
                filtered_revit_files_intermediate.append(file)
        return filtered_revit_files_intermediate
    
    def TextBox_Filter_TextChanged(self):
        """
        Event handler for text changed in TextBox.
        """

        # break up filter text into individual rules if required
        if(";" in self.filter_text):
            self.filter_rules = [rule.strip() for rule in self.filter_text.split(";")]
        else:
            # just a single filter applied
            self.filter_rules=[self.filter_text]
            
        # clear the previous version of the filtered list
        self.filtered_revit_files.Clear()
        # set up a python vanilla list to contain filtered elements
        filtered_revit_files_intermediate = []
        
        # filter file data depending on filter type
        if(self.filter_is_and):
            filtered_revit_files_intermediate = self._apply_and_to_file_data()
        else:
            filtered_revit_files_intermediate = self._apply_or_to_file_data()
       
        # transfer files from intermediate list to List[ObservableCollection]
        self.convert_list_to_observable_collection(
            filtered_revit_files_intermediate, self.filtered_revit_files
        )

    def _get_files(self):
        """
        Get initial, unfiltered file list
        """
        new_files = get_revit_files_for_processing(
            self.source_path, self.include_sub_dirs, self.source_file_extension
        )

        # convert past in files list
        self.convert_list_to_observable_collection(new_files, self.revit_files)
        
    def text_box_source_path_change(self):
        """
        when the source directory is changes
        """

        self._get_files()
        # update filtered list
        self.TextBox_Filter_TextChanged()

    def button_ok(self):
        """
        Ok button event handler.

        Gets the selected rows of files and adds them to .selectedFiles property
        Sets the dialog result value to True.

        :param sender: _description_
        :type sender: _type_
        :param EventArgs: _description_
        :type EventArgs: _type_
        """

        self.DialogResult = True
        self.selected_files.Clear()  # Clear existing items if any
        # this is also a hack:
        # file item has a  property .is_selected which is bound to IsSelected Row style in XAML file
        # a work around to get all selected rows from the data grid view
        for file_item in self.revit_files:
            if file_item.is_selected:
                self.selected_files.Add(file_item)

        self.window.Close()

    def button_cancel(self):
        """
        Cancel button event handler.
        Sets the dialogue result value to False.

        :param sender: _description_
        :type sender: _type_
        :param EventArgs: _description_
        :type EventArgs: _type_
        """

        self.DialogResult = False
        self.window.Close()
